from collections.abc import Callable
from crystal.url_input import cleaned_url_is_at_site_root, UrlCleaner
from crystal.util.bulkheads import capture_crashes_to_stderr
from crystal.util.ellipsis import Ellipsis, EllipsisType
from crystal.util.test_mode import tests_are_running
from crystal.util.wx_bind import bind
from crystal.util.wx_dialog import (
    CreateButtonSizer, position_dialog_initially, ShowModal,
)
from crystal.util.wx_static_box_sizer import wrap_static_box_sizer_child
from crystal.util.xos import is_linux, is_windows, is_wx_gtk
from crystal.util.xthreading import fg_affinity, fg_call_later
from typing import Literal, Optional, Union
import wx

ChangePrefixCommand = Union[
    # Set prefix
    tuple[Literal['domain', 'directory'], str],
    # Clear prefix
    None,
    # Leave prefix unchanged
    EllipsisType,
]


_WINDOW_INNER_PADDING = 10
_FORM_LABEL_INPUT_SPACING = 5
_FORM_ROW_SPACING = 10
_BESIDE_OPTIONS_PADDING = 10

_OPTIONS_SHOWN_LABEL = 'Basic Options'
_OPTIONS_NOT_SHOWN_LABEL = 'Advanced Options'


class NewRootUrlDialog:
    _ID_SET_DOMAIN_PREFIX = 101
    _ID_SET_DIRECTORY_PREFIX = 102
    
    _ID_DOWNLOAD_IMMEDIATELY = 111
    _ID_CREATE_GROUP = 112
    
    _INITIAL_URL_WIDTH = 400  # in pixels
    _FIELD_TO_SPINNER_MARGIN = 5
    
    # NOTE: Only changed when tests are running
    _last_opened: 'Optional[NewRootUrlDialog]'=None
    
    _url_field: wx.TextCtrl
    _name_field: wx.TextCtrl
    _ok_button: wx.Button
    _cancel_button: wx.Button
    _options_button: wx.Button
    
    _download_immediately_checkbox: wx.CheckBox
    _create_group_checkbox: wx.CheckBox
    
    _set_as_default_domain_checkbox: wx.CheckBox
    _set_as_default_directory_checkbox: wx.CheckBox
    
    # === Init ===
    
    def __init__(self,
            parent: wx.Window,
            on_finish: Callable[[str, str, ChangePrefixCommand, bool, bool], None],
            url_exists_func: Callable[[str], bool],
            initial_url: str='',
            initial_name: str='',
            initial_set_as_default_domain: bool=False,
            initial_set_as_default_directory: bool=False,
            allow_set_as_default_domain_or_directory: bool=True,
            is_edit: bool=False,
            ) -> None:
        """
        Arguments:
        * parent -- parent wx.Window that this dialog is attached to.
        * on_finish -- called when OK pressed on dialog.
        * initial_url -- overrides the initial URL displayed.
        """
        self._on_finish = on_finish
        self._url_exists_func = url_exists_func
        self._did_own_prefix = (
            is_edit and
            (initial_set_as_default_domain or initial_set_as_default_directory)
        )
        self._is_edit = is_edit
        
        self._url_field_focused = False
        self._last_cleaned_url = None  # type: Optional[str]
        self._url_cleaner = None  # type: Optional[UrlCleaner]
        self._was_ok_pressed = False
        self._is_destroying_or_destroyed = False
        
        dialog = self.dialog = wx.Dialog(
            parent,
            title=('New Root URL' if not is_edit else 'Edit Root URL'),
            name='cr-new-root-url-dialog',
            style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)
        dialog_sizer = wx.BoxSizer(wx.VERTICAL)
        dialog.SetSizer(dialog_sizer)
        bind(dialog, wx.EVT_BUTTON, self._on_button)
        bind(dialog, wx.EVT_CLOSE, self._on_close)
        bind(dialog, wx.EVT_WINDOW_DESTROY, self._on_destroyed)
        
        dialog_sizer.Add(
            self._create_fields(dialog, initial_url, initial_name, is_edit),
            flag=wx.EXPAND|wx.ALL,
            border=_WINDOW_INNER_PADDING)
        
        if not is_edit:
            new_options_sizer = self._create_new_options(dialog)
            dialog_sizer.Add(
                new_options_sizer,
                flag=wx.EXPAND|wx.LEFT|wx.RIGHT|wx.BOTTOM,
                border=self._assert_same(
                    _WINDOW_INNER_PADDING,  # wx.LEFT|wx.RIGHT
                    _BESIDE_OPTIONS_PADDING,  # wx.BOTTOM
                ))
        
        self._advanced_options_sizer = self._create_advanced_options(
            dialog,
            initial_set_as_default_domain, initial_set_as_default_directory,
            allow_set_as_default_domain_or_directory)
        dialog_sizer.Add(
            self._advanced_options_sizer,
            flag=wx.EXPAND|wx.LEFT|wx.RIGHT|wx.BOTTOM,
            border=self._assert_same(
                _WINDOW_INNER_PADDING,  # wx.LEFT|wx.RIGHT
                _BESIDE_OPTIONS_PADDING,  # wx.BOTTOM
            ))
    
        dialog_sizer.Add(
            self._create_buttons(dialog, is_edit),
            flag=wx.EXPAND|wx.BOTTOM,
            border=_WINDOW_INNER_PADDING)
        
        self._update_ok_enabled()
        
        if not fields_hide_hint_when_focused():
            # Initialize focus
            if not is_edit:
                self._url_field.SetFocus()
            else:
                self._name_field.SetFocus()
        
        position_dialog_initially(dialog)
        # TODO: Verify that the wxGTK-specific logic here is actually necessary
        if not is_wx_gtk():
            dialog.Fit()
            self._on_advanced_options_toggle()  # collapse options initially
            dialog.Show(True)
        else:
            dialog.Show(True)
            dialog.Fit()  # NOTE: Must Fit() after Show() here so that wxGTK actually fits correctly
            self._on_advanced_options_toggle()  # collapse options initially
        
        dialog.MinSize = dialog.Size
        # TODO: Clamp height to fixed value, but still allow
        #       dialog to change height when options shown/hidden
        dialog.MaxSize = wx.Size(wx.DefaultCoord, wx.DefaultCoord)
        
        # Export reference to self, if running tests
        if tests_are_running():
            NewRootUrlDialog._last_opened = self
    
    @staticmethod
    def _assert_same(v1: int, v2: int) -> int:
        assert v1 == v2
        return v1
    
    def _create_fields(self, parent: wx.Window, initial_url: str, initial_name: str, is_edit: bool) -> wx.Sizer:
        fields_sizer = wx.FlexGridSizer(rows=2, cols=2,
            vgap=_FORM_ROW_SPACING, hgap=_FORM_LABEL_INPUT_SPACING)
        fields_sizer.AddGrowableCol(1)
        if True:
            url_label = wx.StaticText(parent, label='URL:', style=wx.ALIGN_RIGHT)
            url_label.Font = url_label.Font.Bold()  # mark as required
            fields_sizer.Add(url_label, flag=wx.EXPAND)
            
            url_field_and_spinner = wx.BoxSizer(wx.HORIZONTAL)
            spinner_diameter: int
            if True:
                self._url_field = wx.TextCtrl(
                    parent, value=initial_url,
                    size=(self._INITIAL_URL_WIDTH, wx.DefaultCoord),
                    name='cr-new-root-url-dialog__url-field')
                self._url_field.Hint = 'https://example.com/'
                self._url_field.SetSelection(-1, -1)  # select all upon focus
                self._url_field.Enabled = not is_edit
                bind(self._url_field, wx.EVT_TEXT, self._on_url_field_changed)
                bind(self._url_field, wx.EVT_SET_FOCUS, self._on_url_field_focus)
                bind(self._url_field, wx.EVT_KILL_FOCUS, self._on_url_field_blur)
                url_field_and_spinner.Add(self._url_field, proportion=1, flag=wx.EXPAND)
                
                spinner_diameter = self._url_field.Size.Height
                self.url_cleaner_spinner = wx.ActivityIndicator(
                    parent,
                    size=wx.Size(spinner_diameter, spinner_diameter),
                    name='cr-new-root-url-dialog__url-cleaner-spinner')
                self.url_cleaner_spinner.Hide()
                url_field_and_spinner.Add(
                    self.url_cleaner_spinner,
                    flag=wx.LEFT|wx.RESERVE_SPACE_EVEN_IF_HIDDEN,
                    border=self._FIELD_TO_SPINNER_MARGIN)
            fields_sizer.Add(url_field_and_spinner, flag=wx.EXPAND)
            
            fields_sizer.Add(wx.StaticText(parent, label='Name:', style=wx.ALIGN_RIGHT), flag=wx.EXPAND)
            
            name_field_and_space = wx.BoxSizer(wx.HORIZONTAL)
            if True:
                self._name_field = wx.TextCtrl(
                    parent, value=initial_name,
                    name='cr-new-root-url-dialog__name-field')
                self._name_field.Hint = 'Home'
                self._name_field.SetSelection(-1, -1)  # select all upon focus
                name_field_and_space.Add(self._name_field, proportion=1, flag=wx.EXPAND)
                
                #name_field_and_space.Add(
                #    wx.Size(spinner_diameter, spinner_diameter),
                #    flag=wx.LEFT,
                #    border=self._FIELD_TO_SPINNER_MARGIN)
            fields_sizer.Add(name_field_and_space, flag=wx.EXPAND)
        
        return fields_sizer
    
    def _create_new_options(self, parent: wx.Window) -> wx.StaticBoxSizer:
        options_sizer = wx.StaticBoxSizer(wx.VERTICAL, parent, label='New URL Options')
        options_sizer.Add(
            wrap_static_box_sizer_child(
                self._create_new_options_content(
                    options_sizer.GetStaticBox())),
            flag=wx.EXPAND)
        return options_sizer
    
    def _create_new_options_content(self, parent: wx.Window) -> wx.Sizer:
        options_sizer = wx.BoxSizer(wx.VERTICAL)
        
        self._download_immediately_checkbox = wx.CheckBox(parent,
            id=self._ID_DOWNLOAD_IMMEDIATELY,
            label='Download Immediately',
            name='cr-new-root-url-dialog__download-immediately-checkbox')
        self._download_immediately_checkbox.Value = True
        options_sizer.Add(self._download_immediately_checkbox,
            flag=wx.BOTTOM,
            border=_FORM_LABEL_INPUT_SPACING)
        
        self._create_group_checkbox = wx.CheckBox(parent,
            id=self._ID_CREATE_GROUP,
            label='Create Group to Download Entire Site',
            name='cr-new-root-url-dialog__create-group-checkbox')
        self._create_group_checkbox.Value = False
        options_sizer.Add(self._create_group_checkbox,
            flag=wx.BOTTOM,
            border=5)
        
        only_simple_sites_label = wx.StaticText(parent, label='')
        only_simple_sites_label.ForegroundColour = \
            wx.Colour(102, 102, 102)  # gray
        only_simple_sites_label.Font = only_simple_sites_label.Font.MakeSmaller()
        # NOTE: Underline formatting is not supported on Windows,
        #       even when using GenStaticText as an alternative to wx.StaticText
        only_simple_sites_label.SetLabelMarkup(
            'Only recommended for downloading <u>simple sites</u>')
        # NOTE: Altering the label's cursor on wxGTK also seems to set
        #       the cursor of its enclosing Sizer as a side effect,
        #       which is confusing. So don't set any cursor on wxGTK.
        if not is_wx_gtk():
            only_simple_sites_label.Cursor = wx.Cursor(wx.CURSOR_QUESTION_ARROW)
        only_simple_sites_label.ToolTip = (
            'A simple website is created by a single person only, '
            'may contain text and images but not video, '
            'and does not requiring logging in to view its content.'
        )
        options_sizer.Add(only_simple_sites_label,
            flag=wx.LEFT,
            border=(
                17 if is_windows() else
                24 if is_linux() else
                20  # macOS
            ))
        
        bind(parent, wx.EVT_CHECKBOX, self._on_checkbox)
        
        self._update_create_group_enabled()
        self._update_download_immediately_label()
        
        return options_sizer
    
    def _create_advanced_options(self, parent: wx.Window, *args, **kwargs) -> wx.StaticBoxSizer:
        options_sizer = wx.StaticBoxSizer(wx.VERTICAL, parent, label='Advanced Options')
        options_sizer.Add(
            wrap_static_box_sizer_child(
                self._create_advanced_options_content(
                    options_sizer.GetStaticBox(),
                    *args, **kwargs)),
            flag=wx.EXPAND)
        return options_sizer
    
    def _create_advanced_options_content(self,
            parent: wx.Window,
            initial_set_as_default_domain: bool,
            initial_set_as_default_directory: bool,
            allow_set_as_default_domain_or_directory: bool,
            ) -> wx.Sizer:
        options_sizer = wx.BoxSizer(wx.VERTICAL)
        
        self._set_as_default_domain_checkbox = wx.CheckBox(parent,
            id=self._ID_SET_DOMAIN_PREFIX,
            label='Set As Default Domain',
            name='cr-new-root-url-dialog__set-as-default-domain-checkbox')
        self._set_as_default_domain_checkbox.Value = initial_set_as_default_domain
        self._set_as_default_domain_checkbox.Enabled = allow_set_as_default_domain_or_directory
        options_sizer.Add(self._set_as_default_domain_checkbox,
            flag=wx.BOTTOM,
            border=_FORM_LABEL_INPUT_SPACING)
        
        self._set_as_default_directory_checkbox = wx.CheckBox(parent,
            id=self._ID_SET_DIRECTORY_PREFIX,
            label='Set As Default Directory',
            name='cr-new-root-url-dialog__set-as-default-directory-checkbox')
        self._set_as_default_directory_checkbox.Value = initial_set_as_default_directory
        self._set_as_default_directory_checkbox.Enabled = allow_set_as_default_domain_or_directory
        options_sizer.Add(self._set_as_default_directory_checkbox)
        
        bind(parent, wx.EVT_CHECKBOX, self._on_checkbox)
        
        return options_sizer
    
    def _create_buttons(self, parent: wx.Window, is_edit: bool) -> wx.Sizer:
        ok_button_id = (wx.ID_NEW if not is_edit else wx.ID_SAVE)
        
        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        
        self._options_button = wx.Button(parent, wx.ID_MORE, _OPTIONS_SHOWN_LABEL)
        button_sizer.Add(
            self._options_button,
            flag=wx.CENTER|wx.LEFT,
            border=_WINDOW_INNER_PADDING)
        
        button_sizer.AddStretchSpacer()
        button_sizer.Add(CreateButtonSizer(parent, ok_button_id, wx.ID_CANCEL), flag=wx.CENTER)
        
        self._ok_button = parent.FindWindow(id=ok_button_id)
        self._cancel_button = parent.FindWindow(id=wx.ID_CANCEL)
        
        return button_sizer
    
    # === URL Cleaning ===
    
    @fg_affinity
    def _on_url_cleaner_running_changed(self, running: bool) -> None:
        if self._is_destroying_or_destroyed:
            return
        
        if running:
            self.url_cleaner_spinner.Show()
            self.url_cleaner_spinner.Start()
        else:
            self.url_cleaner_spinner.Hide()
            self.url_cleaner_spinner.Stop()
            
            self._url_cleaner = None
            
            if self._was_ok_pressed:
                # Continue executing the OK action
                self._on_ok()

    @fg_affinity
    def _set_cleaned_url(self, cleaned_url: str) -> None:
        if self._is_destroying_or_destroyed:
            return
        
        self._last_cleaned_url = cleaned_url
        if self._url_field.Value != cleaned_url:
            self._url_field.Value = cleaned_url
    
    # === Events ===
    
    @fg_affinity
    def _on_url_field_changed(self, event=None) -> None:
        self._update_create_group_enabled()
        self._update_ok_enabled()
    
    # NOTE: Focus event can be called multiple times without an intermediate blur event
    @fg_affinity
    def _on_url_field_focus(self, event: wx.FocusEvent) -> None:
        # NOTE: Cannot use _url_field.HasFocus() because doesn't work in automated tests
        if self._url_field_focused:
            # Already did focus action
            return
        self._url_field_focused = True
        
        # If field still focused after a tick,
        # stop cleaning any old URL input
        @capture_crashes_to_stderr  # no good location in UI to route crashes too
        def fg_task() -> None:
            # NOTE: Cannot use _url_field.HasFocus() because doesn't work in automated tests
            if not self._url_field_focused:
                return
            
            # Stop cleaning any old URL input
            if self._url_cleaner is not None:
                self._url_cleaner.cancel()
                self._url_cleaner = None
        fg_call_later(fg_task, force_later=True)
        
        # Continue processing event in the normal fashion
        event.Skip()
    
    # NOTE: Blur event can be called multiple times without an intermediate focus event
    # NOTE: Blur event can be called while dialog is being destroyed
    @fg_affinity
    def _on_url_field_blur(self, event: wx.FocusEvent | None=None) -> None:
        if not self._url_field_focused:
            # Already did blur action
            return
        self._url_field_focused = False
        
        if self._is_destroying_or_destroyed:
            return
        
        # Start cleaning the new URL input
        url_input = self._url_field.Value
        if url_input == self._last_cleaned_url:
            # URL is already clean
            pass
        else:
            if self._url_cleaner is None or self._url_cleaner.url_input != url_input:
                if self._url_cleaner is not None:
                    self._url_cleaner.cancel()
                    self._url_cleaner = None
                
                self._url_cleaner = UrlCleaner(
                    url_input,
                    self._on_url_cleaner_running_changed,
                    self._set_cleaned_url)
                self._url_cleaner.start()
                # (NOTE: self._url_cleaner may be None now)
        
        # Continue processing event in the normal fashion
        if event is not None:
            event.Skip()
    
    @fg_affinity
    def _on_button(self, event: wx.CommandEvent) -> None:
        btn_id = event.GetEventObject().GetId()
        if btn_id in (wx.ID_NEW, wx.ID_SAVE):
            self._on_ok(event)
        elif btn_id == wx.ID_CANCEL:
            self._on_cancel(event)
        elif btn_id == wx.ID_MORE:
            self._on_advanced_options_toggle()
    
    @fg_affinity
    def _on_checkbox(self, event: wx.CommandEvent) -> None:
        checkbox_id = event.GetEventObject().GetId()
        if checkbox_id == self._ID_CREATE_GROUP:
            self._update_download_immediately_label()
        elif checkbox_id == self._ID_SET_DOMAIN_PREFIX:
            if self._set_as_default_domain_checkbox.Value:
                self._set_as_default_directory_checkbox.Value = False
        elif checkbox_id == self._ID_SET_DIRECTORY_PREFIX:
            if self._set_as_default_directory_checkbox.Value:
                self._set_as_default_domain_checkbox.Value = False
    
    @fg_affinity
    def _on_close(self, event: wx.CloseEvent) -> None:
        self._on_cancel(event)
    
    @fg_affinity
    def _on_ok(self, event: wx.CommandEvent | None=None) -> None:
        # Blur text fields so that they save their contents
        self._on_url_field_blur()
        
        # If URL input is being cleaned, wait for it to finish before continuing
        if self._url_cleaner is not None:
            self._url_field.Enabled = False
            self._name_field.Enabled = False
            self._ok_button.Enabled = False
            assert self._cancel_button.Enabled == True
            
            self._was_ok_pressed = True
            return
        
        name = self._name_field.Value
        url = self._url_field.Value
        if not self._is_edit and self._url_exists_func(url):
            dialog = wx.MessageDialog(
                self.dialog,
                message='That root URL already exists in the project.',
                caption='Root URL Exists',
                style=wx.OK,
            )
            dialog.Name = 'cr-root-url-exists'
            position_dialog_initially(dialog)
            choice = ShowModal(dialog)
            assert wx.ID_OK == choice
            
            self._url_field.Enabled = True
            self._name_field.Enabled = True
            self._ok_button.Enabled = True
            assert self._cancel_button.Enabled == True
            return
        if self._set_as_default_domain_checkbox.Value:
            change_prefix_command = ('domain', url)  # type: ChangePrefixCommand
        elif self._set_as_default_directory_checkbox.Value:
            change_prefix_command = ('directory', url)
        else:
            change_prefix_command = None if self._did_own_prefix else Ellipsis
        if self._is_edit:
            download_immediately = False
            create_group = False
        else:
            download_immediately = self._download_immediately_checkbox.Value
            create_group = self._create_group_checkbox.Value
        self._on_finish(
            name, url, change_prefix_command,
            download_immediately, create_group)
        
        self._destroy()
    
    @fg_affinity
    def _on_cancel(self, event: wx.CommandEvent) -> None:
        # Stop cleaning any old URL input
        if self._url_cleaner is not None:
            self._url_cleaner.cancel()
            self._url_cleaner = None
        
        self._destroy()
    
    @fg_affinity
    def _on_advanced_options_toggle(self) -> None:
        options = self._advanced_options_sizer.GetStaticBox()
        if options.Shown:
            # Hide
            self._options_button.Label = _OPTIONS_NOT_SHOWN_LABEL
            
            options_height = options.Size.Height + _BESIDE_OPTIONS_PADDING
            options.Shown = False
            self.dialog.SetSize(
                x=wx.DefaultCoord,
                y=wx.DefaultCoord,
                width=wx.DefaultCoord,
                height=self.dialog.Size.Height - options_height,
                sizeFlags=wx.SIZE_USE_EXISTING)
        else:
            # Show
            self._options_button.Label = _OPTIONS_SHOWN_LABEL
            
            options.Shown = True
            options_height = options.Size.Height + _BESIDE_OPTIONS_PADDING
            self.dialog.SetSize(
                x=wx.DefaultCoord,
                y=wx.DefaultCoord,
                width=wx.DefaultCoord,
                height=self.dialog.Size.Height + options_height,
                sizeFlags=wx.SIZE_USE_EXISTING)
    
    @fg_affinity
    def _destroy(self) -> None:
        self._is_destroying_or_destroyed = True
        self.dialog.Destroy()
    
    @fg_affinity
    def _on_destroyed(self, event) -> None:
        self._is_destroying_or_destroyed = True
    
    # === Updates ===
    
    def _update_create_group_enabled(self) -> None:
        url = self._url_field.Value
        self._create_group_checkbox.Enabled = (
            url == '' or cleaned_url_is_at_site_root(url)
        )
        
        self._update_download_immediately_label()
    
    def _update_download_immediately_label(self) -> None:
        self._download_immediately_checkbox.Label = (
            'Download Site Immediately' if (
                self._create_group_checkbox.Enabled and
                self._create_group_checkbox.Value
            ) else 'Download URL Immediately'
        )
    
    def _update_ok_enabled(self) -> None:
        self._ok_button.Enabled = (self._url_field.Value != '')


def fields_hide_hint_when_focused() -> bool:
    return is_wx_gtk()



"""
Tests the --install-to-desktop command line option on Linux,
implemented by the install.py module.
"""
from unittest import skip

# === Not Linux ===

@skip('not yet automated')
def test_given_not_linux_when_install_to_desktop_then_exits_with_error_message(self) -> None:
    pass


# === Non-Root vs. Root User ===

@skip('not yet automated')
def test_given_linux_and_non_root_user_is_using_sudo_when_install_to_desktop_then_exits_with_error_message(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_non_root_user_is_not_using_sudo_when_install_to_desktop_then_does_not_exit_with_error_message(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_is_root_user_when_install_to_desktop_then_does_not_exit_with_error_message(self) -> None:
    pass


# === Installs to Desktop ===

@skip('not yet automated')
def test_given_linux_when_install_to_desktop_then_crystal_app_appears_on_desktop_with_custom_icon_and_can_be_double_clicked_to_open_crystal(self) -> None:
    pass


# === Installs to Application Launcher ===

@skip('not yet automated')
def test_given_linux_and_gnome_desktop_environment_when_install_to_desktop_then_crystal_app_appears_in_application_launcher_menus(self) -> None:
    pass


@skip('not yet automated')
def test_given_linux_and_kde_desktop_environment_when_install_to_desktop_then_crystal_app_appears_in_application_launcher_menus(self) -> None:
    pass



from urllib.parse import urlparse, urlunparse


def get_url_domain_prefix_for(url_or_url_prefix: str) -> str | None:
    """
    Given a URL or URL prefix, returns the URL for its enclosing domain,
    with no trailing slash.
    """
    url_components = urlparse(url_or_url_prefix)
    if url_components.scheme not in ('http', 'https'):
        return None
    
    return urlunparse(url_components._replace(
        path='',
        params='',
        query='',
        fragment='',
    ))

def get_url_directory_prefix_for(url_or_url_prefix: str) -> str | None:
    """
    Given a URL or URL prefix, returns the URL for its enclosing directory,
    with no trailing slash.
    """
    url_components = urlparse(url_or_url_prefix)
    if url_components.scheme not in ('http', 'https'):
        return None
    
    # If URL path contains slash, chop last slash and everything following it
    path = url_components.path
    if '/' in path:
        new_path = path[:path.rindex('/')]
    else:
        new_path = path
    
    return urlunparse(url_components._replace(path=new_path))



"""
Facade for working with wx.TreeCtrl.

This abstraction provides:
* tree nodes that can be manipulated before being added to a tree
* delegate-style event handling on tree nodes
* access to the underlying "peer" objects (i.e. wx.TreeCtrl, tree item index)
"""

from __future__ import annotations

from collections.abc import Callable, Container
from crystal.progress import OpenProjectProgressListener
from crystal.util.bulkheads import run_bulkhead_call
from crystal.util.wx_bind import bind
from crystal.util.wx_error import (
    IGNORE_USE_AFTER_FREE, is_wrapped_object_deleted_error, WindowDeletedError,
    wrapped_object_deleted_error_ignored, wrapped_object_deleted_error_raising,
)
from crystal.util.xthreading import fg_affinity
from typing import cast, Dict, List, NewType, NoReturn, Optional
import wx

IconSet = tuple[tuple[wx.TreeItemIcon, wx.Bitmap], ...]
ImageIndex = NewType('ImageIndex', int)

_DEFAULT_TREE_ICON_SIZE = (16,16)

_DEFAULT_FOLDER_ICON_SET_CACHED = None
def DEFAULT_FOLDER_ICON_SET() -> IconSet:
    global _DEFAULT_FOLDER_ICON_SET_CACHED  # necessary to write to a module global
    if not _DEFAULT_FOLDER_ICON_SET_CACHED:
        _DEFAULT_FOLDER_ICON_SET_CACHED = (
            (wx.TreeItemIcon_Normal,   wx.ArtProvider.GetBitmap(wx.ART_FOLDER,      wx.ART_OTHER, _DEFAULT_TREE_ICON_SIZE)),
            (wx.TreeItemIcon_Expanded, wx.ArtProvider.GetBitmap(wx.ART_FILE_OPEN,   wx.ART_OTHER, _DEFAULT_TREE_ICON_SIZE)),
        )
    return _DEFAULT_FOLDER_ICON_SET_CACHED

_DEFAULT_FILE_ICON_SET_CACHED = None
def _DEFAULT_FILE_ICON_SET() -> IconSet:
    global _DEFAULT_FILE_ICON_SET_CACHED    # necessary to write to a module global
    if not _DEFAULT_FILE_ICON_SET_CACHED:
        _DEFAULT_FILE_ICON_SET_CACHED = (
            (wx.TreeItemIcon_Normal,   wx.ArtProvider.GetBitmap(wx.ART_NORMAL_FILE, wx.ART_OTHER, _DEFAULT_TREE_ICON_SIZE)),
        )
    return _DEFAULT_FILE_ICON_SET_CACHED

# Maps wx.EVT_TREE_ITEM_* events to names of methods on `NodeView.delegate`
# that will be called (if they exist) upon the reception of such an event.
_EVENT_TYPE_2_DELEGATE_CALLABLE_ATTR = {
    wx.EVT_TREE_ITEM_EXPANDED: 'on_expanded',
    wx.EVT_TREE_ITEM_RIGHT_CLICK: 'on_right_click',
    # TODO: Consider adding support for additional wx.EVT_TREE_ITEM_* event types
}
_EVENT_TYPE_ID_2_DELEGATE_CALLABLE_ATTR = dict(zip(
    [et.typeId for et in _EVENT_TYPE_2_DELEGATE_CALLABLE_ATTR],
    _EVENT_TYPE_2_DELEGATE_CALLABLE_ATTR.values()
))


class TreeView:
    """
    Displays a tree of nodes.
    
    Acts as a facade for manipulating an underlying wx.TreeCtrl.
    For advanced customization, this wx.TreeCtrl may be accessed through the `peer` attribute.
    
    Automatically creates a root NodeView (accessible via the `root` attribute),
    which will not be displayed 
    """
    
    def __init__(self, parent_peer: wx.Window, *, name: str | None=None) -> None:
        self.delegate = None  # type: object
        self.peer = _OrderedTreeCtrl(
            parent_peer,
            style=wx.TR_DEFAULT_STYLE|wx.TR_HIDE_ROOT,
            **(
                dict(name=name)
                if name is not None else
                dict()
            ))  # type: wx.TreeCtrl
        
        # Setup node image registration
        # NOTE: In wxPython 4.2.0, wx.Bitmap icons will be superceded by wx.BitmapBundle,
        #       and wx.ImageList will be superceded by a plain list of wx.BitmapBundles.
        self._bitmap_2_image_id = dict()  # type: Dict[wx.Bitmap, ImageIndex]
        self._tree_imagelist = wx.ImageList(*_DEFAULT_TREE_ICON_SIZE)
        self.peer.AssignImageList(self._tree_imagelist)
        
        # Create root node's view
        self._root_peer = NodeViewPeer(self, self.peer.AddRoot(''))
        self.root = NodeView()
        
        # Listen for events on peer
        for event_type in _EVENT_TYPE_2_DELEGATE_CALLABLE_ATTR:
            bind(self.peer, event_type, self._dispatch_event, self.peer)
    
    # === Properties ===
    
    def _get_root(self) -> NodeView:
        return self._root
    def _set_root(self, value: NodeView) -> None:
        self._root = value
        self._root._attach(self._root_peer)
    root = property(_get_root, _set_root)
    
    # TODO: Rename: selection
    @property
    def selected_node(self) -> NodeView | None:
        return self.selected_node_in(self.peer)
    
    # TODO: Rename: selection_in
    @staticmethod
    def selected_node_in(tree_peer: wx.TreeCtrl) -> NodeView | None:
        try:
            selected_node_id = tree_peer.GetSelection()
        except Exception as e:
            if is_wrapped_object_deleted_error(e):
                return None
            else:
                raise
        if not selected_node_id.IsOk():
            return None
        return tree_peer.GetItemData(selected_node_id)
    
    def get_image_id_for_bitmap(self, bitmap: wx.Bitmap) -> ImageIndex:
        """
        Given a wx.Bitmap, returns an image ID suitable to use as an node icon.
        Calling this multiple times with the same wx.Bitmap will return the same image ID.
        """
        if bitmap in self._bitmap_2_image_id:
            image_id = self._bitmap_2_image_id[bitmap]
        else:
            image_id = self._tree_imagelist.Add(bitmap)
            self._bitmap_2_image_id[bitmap] = ImageIndex(image_id)
        return image_id
    
    # === Operations ===
    
    def expand(self, node_view):
        self.peer.Expand(node_view.peer.node_id)
    
    # Notified when any interesting event occurs on the peer
    def _dispatch_event(self, event: wx.TreeEvent) -> None:
        node_id = event.GetItem()
        node_view = self.peer.GetItemData(node_id)  # type: NodeView
        
        # Dispatch event to the node
        node_view._dispatch_event(event)
        
        # Dispatch event to my delegate
        if self.delegate:
            event_type_id = event.GetEventType()
            delegate_callable_attr = _EVENT_TYPE_ID_2_DELEGATE_CALLABLE_ATTR.get(event_type_id, None)
            if delegate_callable_attr and hasattr(self.delegate, delegate_callable_attr):
                run_bulkhead_call(
                    getattr(self.delegate, delegate_callable_attr),
                    event,
                    node_view
                )
    
    def dispose(self) -> None:
        self.delegate = None
        self.root.dispose()


class _OrderedTreeCtrl(wx.TreeCtrl):
    def OnCompareItems(self, item1: wx.TreeItemId, item2: wx.TreeItemId) -> int:
        (item1_view, item2_view) = (self.GetItemData(item1), self.GetItemData(item2))
        assert isinstance(item1_view, NodeView) and isinstance(item2_view, NodeView)
        (order_index_1, order_index_2) = (
            getattr(item1_view, '_order_index', None),
            getattr(item2_view, '_order_index', None)
        )
        if order_index_1 is None and order_index_2 is None:
            return 0
        assert isinstance(order_index_1, int) and isinstance(order_index_2, int)
        return order_index_1 - order_index_2


class NodeView:
    """
    Node that is (or will be) in a TreeView.
    
    Acts as a facade for manipulating a wxTreeItemId in a wxTreeCtrl. Allows modifications even
    if the underlying wxTreeItemId doesn't yet exist. For advanced customization, the wxTreeItemId
    and wxTreeCtrl may be accessed through the `peer` attribute (which is a `NodeViewPeer`)).
    
    To receive events that occur on a NodeView, assign an object to the `delegate` attribute.
    * For each event of interest, this object should implement methods of the signature:
          def on_eventname(self, event)
    * The `event` object passed to this method is a wx.Event object that can be inspected for more
      information about the event.
    * The full list of supported event names is given by
      `_EVENT_TYPE_ID_2_DELEGATE_CALLABLE_ATTR.values()`.
    """
    DEFAULT_TEXT_COLOR = wx.Colour(0, 0, 0)  # black
    
    # Optimize per-instance memory use, since there may be very many NodeView objects
    __slots__ = (
        'delegate',
        'peer',
        '_title',
        '_text_color',
        '_bold',
        '_expandable',
        '_icon_set_func',
        '_icon_set',
        '_children',
        '_order_index',
    )
    
    def __init__(self) -> None:
        self.delegate = None  # type: object
        self.peer = None  # type: Optional[NodeViewPeer]
        self._title = ''
        self._text_color = None  # type: Optional[wx.Colour]
        self._bold = False
        self._expandable = False
        self._icon_set_func = None  # type: Optional[Callable[[], Optional[IconSet]]]
        self._icon_set = None  # type: Optional[IconSet]
        self._children = []  # type: List[NodeView]
    
    # === Properties ===
    
    def _get_title(self) -> str:
        return self._title
    def _set_title(self, value: str) -> None:
        self._title = value
        if self.peer:
            self.peer.SetItemText(value)
    title = property(_get_title, _set_title)
    
    def _get_text_color(self) -> wx.Colour | None:
        return self._text_color
    def _set_text_color(self, value: wx.Colour | None) -> None:
        self._text_color = value
        if self.peer:
            self.peer.SetItemTextColour(value or self.DEFAULT_TEXT_COLOR)
    text_color = property(_get_text_color, _set_text_color)
    
    def _get_bold(self) -> bool:
        return self._bold
    def _set_bold(self, value: bool) -> None:
        self._bold = value
        if self.peer:
            self.peer.SetItemBold(value)
    bold = property(_get_bold, _set_bold)
    
    def _get_expandable(self) -> bool:
        return self._expandable
    def _set_expandable(self, value: bool) -> None:
        self._expandable = value
        if self.peer:
            self.peer.SetItemHasChildren(value)
            # If using default icon set, force it to update since it depends on the expandable state
            if self.icon_set is None:
                self.icon_set = self.icon_set
    expandable = property(_get_expandable, _set_expandable)
    
    def _get_icon_set(self) -> IconSet | None:
        """
        A sequence of (wx.TreeItemIcon, wx.Bitmap) tuples, specifying the set of icons applicable
        to this node in various states. If None, then a default icon set is used, depending on
        whether this node is expandable.
        """
        if self._icon_set_func is not None:
            self._icon_set = self._icon_set_func()
            self._icon_set_func = None
        return self._icon_set
    def _set_icon_set(self,
            value: (
                IconSet | None |
                Callable[[], IconSet | None]  # deferred value
            )) -> None:
        if callable(value):
            self._icon_set_func = value
            self._icon_set = None
            return
        
        self._icon_set = value
        if self.peer:
            effective_value = (
                value
                if value is not None else (
                    DEFAULT_FOLDER_ICON_SET()
                    if self.expandable 
                    else _DEFAULT_FILE_ICON_SET()
                )
            )  # type: IconSet
            for (which, bitmap) in effective_value:
                self.peer.SetItemImage(self._tree.get_image_id_for_bitmap(bitmap), which)
    icon_set = property(_get_icon_set, _set_icon_set)
    
    def _get_children(self) -> list[NodeView]:
        return self._children
    def _set_children(self, new_children: list[NodeView]) -> None:
        self.set_children(new_children)
    children = cast('List[NodeView]', property(_get_children, _set_children))
    
    def set_children(self,
            new_children: list[NodeView],
            progress_listener: OpenProjectProgressListener | None=None,
            *, _initial: bool=False
            ) -> None:
        """
        Raises:
        * CancelOpenProject
        """
        if progress_listener is not None:
            part_count = sum([len(c.children) for c in new_children])
            progress_listener.creating_entity_tree_nodes(part_count)
        
        old_children = self._children
        self._children = new_children
        if self.peer:
            try:
                if _initial or len(old_children) == 0:
                    # Add initial children
                    for (index, child) in enumerate(new_children):
                        # TODO: Consider storing _order_index in a separate
                        #       child_2_order_index dict rather than annotating
                        #       the child object directly
                        child._order_index = index  # type: ignore[attr-defined]
                        child._attach(NodeViewPeer(self.peer._tree, self.peer.AppendItem('')))
                else:
                    # Replace existing children, preserving old ones that match new ones
                    
                    old_children_set = set(old_children)
                    
                    # 1. Delete some children
                    # 2. Capture selected node if it is in the deleted region
                    if True:
                        children_to_delete = old_children_set - set(new_children)
                        
                        on_selection_deleted = getattr(
                            self.delegate, 'on_selection_deleted', None
                        )  # type: Optional[Callable[[NodeView], Optional[NodeView]]]
                        old_selection_to_retarget = None  # type: Optional[NodeView]
                        if on_selection_deleted is not None:
                            old_selection = TreeView.selected_node_in(self.peer.tree_peer)
                            if old_selection is not None and old_selection.is_descendent_of_any(children_to_delete):
                                old_selection_to_retarget = old_selection  # capture
                        
                        for child in children_to_delete:
                            child._delete_peer_and_descendants()
                    
                    # Add some children
                    children_to_add = [new_child for new_child in new_children if new_child not in old_children_set]
                    for child in children_to_add:
                        child._attach(NodeViewPeer(self.peer._tree, self.peer.AppendItem('')))
                    
                    # Reorder children
                    for (index, child) in enumerate(new_children):
                        child._order_index = index  # type: ignore[attr-defined]
                    self.peer.SortChildren()
                    
                    # If old selected node was in the deleted region, try to
                    # intelligently retarget the selection to something else
                    if on_selection_deleted is not None and old_selection_to_retarget is not None:
                        new_selection = on_selection_deleted(old_selection_to_retarget)
                        if new_selection is None:
                            self.peer.tree_peer.Unselect()
                        else:
                            assert new_selection.peer is not None
                            new_selection.peer.SelectItem()
                    
            except WindowDeletedError:
                if IGNORE_USE_AFTER_FREE:
                    pass
                else:
                    raise
    
    def append_child(self, child: NodeView) -> None:
        # NOTE: The following is equivalent to:
        #           self.children = self.children + [child]
        self._children.append(child)
        if self.peer:
            try:
                child._attach(NodeViewPeer(self.peer._tree, self.peer.AppendItem('')))
            except WindowDeletedError:
                if IGNORE_USE_AFTER_FREE:
                    pass
                else:
                    raise
    
    @property
    def parent(self) -> NodeView | None:
        if not self.peer:
            raise ValueError('Cannot lookup parent when not attached to a tree.')
        parent_treeitemid = self.peer.GetItemParent()
        if not parent_treeitemid.IsOk():
            return None
        return self.peer.tree_peer.GetItemData(parent_treeitemid)
    
    def is_descendent_of_any(self, ancestor_node_views: Container[NodeView]) -> bool:
        cur_node_view = self  # type: Optional[NodeView]
        while cur_node_view is not None:
            if cur_node_view in ancestor_node_views:
                return True
            cur_node_view = cur_node_view.parent
        return False
    
    @property
    def _tree(self) -> TreeView:
        if not self.peer:
            raise ValueError('Not attached to a tree.')
        return self.peer._tree
    
    # === Operations ===
    
    def _attach(self, peer: NodeViewPeer) -> None:
        old_peer = self.peer  # capture
        if old_peer:
            raise ValueError(
                f'Already attached to a different peer: '
                f'old_peer={old_peer!r}, new_peer={peer!r}')
        self.peer = peer
        
        # Enable navigation from peer back to this view
        peer.SetItemData(self)
        
        # Trigger property logic to update peer
        self.title = self.title
        if self._text_color is not None:
            self.text_color = self.text_color
        if self.bold:
            self.bold = self.bold
        self.expandable = self.expandable
        self.icon_set = self.icon_set
        self.set_children(self.children, _initial=True)
    
    def _delete_peer_and_descendants(self) -> None:
        if self.peer is None:
            return
        for child in self._children:
            child._delete_peer_and_descendants()
        self.peer.Delete()
        self.peer = None
    
    # Called when a wx.EVT_TREE_ITEM_* event occurs on this node
    def _dispatch_event(self, event) -> None:
        # Dispatch event to my delegate
        if self.delegate:
            event_type_id = event.GetEventType()
            delegate_callable_attr = _EVENT_TYPE_ID_2_DELEGATE_CALLABLE_ATTR.get(event_type_id, None)
            if delegate_callable_attr and hasattr(self.delegate, delegate_callable_attr):
                run_bulkhead_call(
                    getattr(self.delegate, delegate_callable_attr),
                    event
                )
    
    def dispose(self) -> None:
        self.delegate = None
        self.peer = None
        for c in self._children:
            c.dispose()
        self._children = []
    
    # === Utility ===
    
    def __repr__(self) -> str:
        return f'<NodeView {self.title!r}>'

NULL_NODE_VIEW = NodeView()


class NodeViewPeer(tuple):
    """
    Thin wrapper around a wxPython tree item that makes the underlying API safer to use.
    """
    
    def __new__(cls, tree: TreeView, node_id: wx.TreeItemId):
        return tuple.__new__(cls, (tree, node_id))
    
    # TODO: Only the 'tree_peer' should be stored.
    #       Remove use of this property and update constructor.
    @property
    def _tree(self) -> TreeView:
        return self[0]
    
    @property
    def tree_peer(self) -> wx.TreeCtrl:
        return self._tree.peer
    
    @property
    def node_id(self) -> wx.TreeItemId:
        return self[1]
    
    @fg_affinity
    def SetItemData(self, obj: NodeView) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SetItemData(node_id, obj)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def SetItemText(self, text: str) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SetItemText(node_id, text)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def SetItemTextColour(self, colour: wx.Colour) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SetItemTextColour(node_id, colour)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def SetItemBold(self, bold: bool) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SetItemBold(node_id, bold)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def SetItemHasChildren(self, has: bool) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SetItemHasChildren(node_id, has)
        else:
            self._did_access_not_ok_wx_object()
    
    # TODO: Delete unused method
    @fg_affinity
    def GetFirstChild(self) -> tuple[wx.TreeItemId, object]:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_raising(self._raise_no_longer_exists):
                return self.tree_peer.GetFirstChild(node_id)
        else:
            self._raise_no_longer_exists()
    
    @fg_affinity
    def GetItemParent(self) -> wx.TreeItemId:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_raising(self._raise_no_longer_exists):
                return self.tree_peer.GetItemParent(node_id)
        else:
            self._raise_no_longer_exists()
    
    @fg_affinity
    def AppendItem(self, text: str, *args) -> wx.TreeItemId:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_raising(self._raise_no_longer_exists):
                return self.tree_peer.AppendItem(node_id, text, *args)
        else:
            self._raise_no_longer_exists()
    
    @fg_affinity
    def SetItemImage(self, image: ImageIndex, which: wx.TreeItemIcon) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SetItemImage(node_id, image, which)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def Delete(self) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.Delete(node_id)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def SortChildren(self) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SortChildren(node_id)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def SelectItem(self, select: bool=True) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.SelectItem(node_id, select)
        else:
            self._did_access_not_ok_wx_object()
    
    @fg_affinity
    def IsSelected(self) -> bool:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                return self.tree_peer.IsSelected(node_id)
            return False
        else:
            self._did_access_not_ok_wx_object()
            return False
    
    @fg_affinity
    def Collapse(self) -> None:
        node_id = self.node_id  # cache
        if node_id.IsOk():
            with wrapped_object_deleted_error_ignored():
                self.tree_peer.Collapse(node_id)
        else:
            self._did_access_not_ok_wx_object()
    
    def _raise_no_longer_exists(self) -> NoReturn:
        raise WindowDeletedError('Tree item no longer exists')
    
    def _did_access_not_ok_wx_object(self) -> None:
        if IGNORE_USE_AFTER_FREE:
            pass
        else:
            self._raise_no_longer_exists()



from concurrent.futures import Future
from crystal.model import (
    Project, Resource, ResourceGroup, ResourceRevision, RootResource,
)
from crystal.server import ProjectServer
from crystal.task import (
    ASSUME_RESOURCES_DOWNLOADED_IN_SESSION_WILL_ALWAYS_REMAIN_FRESH,
    DownloadResourceGroupMembersTask, ProjectFreeSpaceTooLowError, Task,
)
from crystal.tests.util.asserts import *
from crystal.tests.util.data import (
    MAX_TIME_TO_DOWNLOAD_404_URL, MAX_TIME_TO_DOWNLOAD_XKCD_HOME_URL_BODY,
)
from crystal.tests.util.downloads import load_children_of_drg_task
from crystal.tests.util.server import fetch_archive_url, served_project
from crystal.tests.util.skip import skipTest
from crystal.tests.util.subtests import awith_subtests, SubtestsContext
from crystal.tests.util.tasks import (
    append_deferred_top_level_tasks, clear_top_level_tasks_on_exit,
    scheduler_disabled, scheduler_thread_context, step_scheduler,
    step_scheduler_until_done,
)
from crystal.tests.util.wait import wait_for
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.tests.util.xthreading import bg_call_and_wait
from crystal.util.progress import ProgressBarCalculator
from crystal.util.xcollections.lazy import AppendableLazySequence
from crystal.util.xthreading import is_foreground_thread
import re
from typing import NamedTuple
from unittest import skip
from unittest.mock import Mock, patch, PropertyMock

# ==============================================================================
# Test: Tasks

# ------------------------------------------------------------------------------
# Test: Task

@awith_subtests
async def test_some_tasks_may_complete_immediately(subtests) -> None:
    assert True == ASSUME_RESOURCES_DOWNLOADED_IN_SESSION_WILL_ALWAYS_REMAIN_FRESH, \
        'Expected optimization to be enabled: ASSUME_RESOURCES_DOWNLOADED_IN_SESSION_WILL_ALWAYS_REMAIN_FRESH'
    
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        if True:
            missing_url = sp.get_request_url('https://example.com/')
            
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            missing_r = Resource(project, missing_url)
            
            with subtests.test(task_type='DownloadResourceTask'):
                # Download the resource
                assert False == missing_r.already_downloaded_this_session
                missing_rr_future = missing_r.download()  # uses DownloadResourceTask
                await wait_for(
                    lambda: missing_rr_future.done() or None,
                    timeout=MAX_TIME_TO_DOWNLOAD_404_URL)
                assert True == missing_r.already_downloaded_this_session
                
                # Download the resource again, and ensure it downloads immediately
                dr_task = missing_r.create_download_task(needs_result=False)  # a DownloadResourceTask
                assert True == dr_task.complete
            
            with subtests.test(task_type='UpdateResourceGroupMembersTask'):
                # Covered by subtest: DownloadResourceGroupTask
                pass
            
            with subtests.test(task_type='DownloadResourceGroupMembersTask'):
                # Covered by subtest: DownloadResourceGroupTask
                pass
            
            with subtests.test(task_type='DownloadResourceGroupTask'):
                comic_rs = [
                    Resource(project, comic_pattern.replace('#', str(ordinal)))
                    for ordinal in [1, 2]
                ]
                
                comic_g = ResourceGroup(project, 'Comic', comic_pattern)
                assert None == comic_g.source
                
                COMIC_G_FINAL_MEMBER_COUNT = 10
                
                # Download the group (and all of its currently known members)
                for r in comic_rs:
                    assert False == r.already_downloaded_this_session
                drg_task = comic_g.create_download_task()  # a DownloadResourceGroupTask
                project.add_task(drg_task)
                await wait_for(
                    lambda: drg_task.complete or None,
                    timeout=(
                        MAX_TIME_TO_DOWNLOAD_404_URL +
                        (MAX_TIME_TO_DOWNLOAD_XKCD_HOME_URL_BODY * COMIC_G_FINAL_MEMBER_COUNT)
                    ))
                assert COMIC_G_FINAL_MEMBER_COUNT == len(comic_g.members)
                for r in comic_rs:
                    assert True == r.already_downloaded_this_session
                
                # Download the group again, and ensure it downloads immediately
                if True:
                    drg_task = comic_g.create_download_task()
                    
                    load_children_of_drg_task(drg_task, task_added_to_project=False)
                    
                    # Precondition
                    # NOTE: The group won't appear to be immediately downloaded yet
                    #       because no code has tried to access the lazily-loaded
                    #       DownloadResourceTask children yet and thus doesn't
                    #       know that all of those children are complete
                    assert (True, 0, False) == (
                        isinstance(drg_task._download_members_task.children, AppendableLazySequence),
                        drg_task._download_members_task.children.cached_prefix_len
                            if isinstance(drg_task._download_members_task.children, AppendableLazySequence)
                            else None,
                        drg_task.complete
                    )
                    
                    # TODO: Disable the scheduler thread before trying to control
                    #       it manually. Currently it's not safe to make the
                    #       scheduler_thread_context() assertion.
                    project.add_task(drg_task)
                    with scheduler_thread_context():
                        await step_scheduler(project, expect_done=True)
                    
                    # Postcondition
                    # NOTE: Adding the DownloadResourceGroupTask to the project's
                    #       task tree will cause the TaskTreeNode to start accessing
                    #       the DownloadResourceTask children because it wants to
                    #       create a paired TaskTreeNode for each such child.
                    #       These accesses cause the DownloadResourceTask children
                    #       to be created and observed as being already complete.
                    #       With all of those children complete the ancestor
                    #       DownloadResourceGroupTask will also be completed.
                    assert (True, COMIC_G_FINAL_MEMBER_COUNT, True) == (
                        isinstance(drg_task._download_members_task.children, AppendableLazySequence),
                        drg_task._download_members_task.children.cached_prefix_len
                            if isinstance(drg_task._download_members_task.children, AppendableLazySequence)
                            else None,
                        drg_task.complete
                    )


def test_given_running_tests_then_uses_extra_listener_assertions() -> None:
    assert True == Task._USE_EXTRA_LISTENER_ASSERTIONS_ALWAYS


# ------------------------------------------------------------------------------
# Test: DownloadResourceTask
#       (+ DownloadResourceBodyTask)
#       (+ ParseResourceRevisionLinks)

# (TODO: Add basic tests)

@skip('not yet automated')
async def test_when_download_resource_then_displays_estimated_time_remaining() -> None:
    pass


def test_format_of_estimated_time_remaining() -> None:
    if ProgressBarCalculator._VERBOSE:
        skipTest('cannot check format when verbose format is being used')
        return
    
    def format_remaining_str(second_count: int) -> str:
        class MockProgressBarCalculator(ProgressBarCalculator):
            pass
        
        pbc = MockProgressBarCalculator(initial=0, total=second_count)
        type(pbc).n = PropertyMock(return_value=0)  # type: ignore[assignment]
        type(pbc).total = PropertyMock(return_value=second_count)
        pbc._rc_n = Mock()
        type(pbc._rc_n).rate = PropertyMock(return_value=1)
        pbc._rc_total = Mock()
        type(pbc._rc_total).rate = PropertyMock(return_value=None)
        
        (remaining_str, time_per_item_str) = pbc.remaining_str_and_time_per_item_str()
        return remaining_str
    
    assertEqual('00:01', format_remaining_str(1))  # 1 second
    assertEqual('01:00', format_remaining_str(60))  # 1 minute
    assertEqual('1:00:00', format_remaining_str(60*60))  # 1 hr
    assertEqual('1d + 0:00:00', format_remaining_str(24*60*60))  # 1 day
    assertEqual('1d + 1:01:01', format_remaining_str(24*60*60 + 60*60 + 60 + 1))


# TODO: Extend test to check how user-facing UI responds to ProjectFreeSpaceTooLowError
#       condition. Currently the UI doesn't handle this condition well and
#       silently fails downloads.
@awith_subtests
async def test_given_project_on_disk_with_low_space_free_when_try_to_download_resource_revision_then_raises_exception(subtests: SubtestsContext) -> None:
    async def try_download_with_disk_usage(du: _DiskUsage, *, expect_failure: bool) -> None:
        assert project is not None
        with patch('shutil.disk_usage', return_value=du):
            r = Resource(project, home_url)
            try:
                rr_future = r.download_body()
                await wait_for(
                    lambda: rr_future.done() or None,
                    timeout=MAX_TIME_TO_DOWNLOAD_XKCD_HOME_URL_BODY)
                try:
                    rr_future.result()
                except ProjectFreeSpaceTooLowError:
                    if expect_failure:
                        pass  # expected
                    else:
                        assert False, f'Expected ProjectFreeSpaceTooLowError but got success when {du}'
                else:
                    if not expect_failure:
                        pass  # expected
                    else:
                        assert False, f'Expected success but got ProjectFreeSpaceTooLowError when {du}'
            finally:
                r.delete()
    
    with subtests.test('given project on small disk and less than 5 percent of disk free'):
        with served_project('testdata_xkcd.crystalproj.zip') as sp:
            # Define URLs
            home_url = sp.get_request_url('https://xkcd.com/')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
                await try_download_with_disk_usage(
                    _DiskUsage(total=100*1024*1024, used=(100-6)*1024*1024, free=6*1024*1024),
                    expect_failure=False)
                
                await try_download_with_disk_usage(
                    _DiskUsage(total=100*1024*1024, used=(100-4)*1024*1024, free=4*1024*1024),
                    expect_failure=True)
    
    with subtests.test('given project on large disk and less than 4 gib of disk free'):
        with served_project('testdata_xkcd.crystalproj.zip') as sp:
            # Define URLs
            home_url = sp.get_request_url('https://xkcd.com/')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
                await try_download_with_disk_usage(
                    _DiskUsage(total=1000*1024*1024*1024, used=(1000-6)*1024*1024*1024, free=6*1024*1024*1024),
                    expect_failure=False)
                
                await try_download_with_disk_usage(
                    _DiskUsage(total=1000*1024*1024*1024, used=(1000-3)*1024*1024*1024, free=3*1024*1024*1024),
                    expect_failure=True)


class _DiskUsage(NamedTuple):
    total: int
    used: int
    free: int


# ------------------------------------------------------------------------------
# Test: DownloadResourceGroupTask
#       (+ UpdateResourceGroupMembersTask)
#       (+ DownloadResourceGroupMembersTask)

# (TODO: Add basic tests)

@skip('not yet automated')
async def test_when_download_resource_group_members_then_displays_estimated_time_remaining() -> None:
    pass


@awith_subtests
async def test_given_download_resource_group_members_when_add_group_member_via_dynamic_browsing_then_new_member_is_queued_for_download(subtests) -> None:
    for children_loaded_before_member_added in [True, False]:
        with subtests.test(children_loaded_before_member_added=children_loaded_before_member_added):
            # NOTE: NOT using the xkcd test data, because I want all xkcd URLs to give HTTP 404
            with scheduler_disabled(), \
                    served_project('testdata_bongo.cat.crystalproj.zip') as sp:
                # Define URLs
                comic1_url = sp.get_request_url('https://xkcd.com/1/')
                comic2_url = sp.get_request_url('https://xkcd.com/2/')
                comic_pattern = sp.get_request_url('https://xkcd.com/#/')
                
                async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
                    with clear_top_level_tasks_on_exit(project):
                        RootResource(project, '', Resource(project, comic1_url))
                        comic_g = ResourceGroup(project, '', comic_pattern)
                        
                        server = ProjectServer(project)
                        try:
                            drg_task = comic_g.create_download_task()
                            project.add_task(drg_task); append_deferred_top_level_tasks(project)
                            
                            drgm_task = drg_task._download_members_task
                            if children_loaded_before_member_added:
                                load_children_of_drg_task(drg_task, scheduler_thread_enabled=False)
                                assertEqual(1, _group_size_in_subtitle(drgm_task))
                            else:
                                assertEqual('Queued', drgm_task.subtitle)
                            
                            # Trigger start of dynamic download of new group member
                            # 
                            # Meanwhile:
                            # - Patch Resource.download() to start download and return an
                            #   already-completed Future with an arbitrary result,
                            #   rather than blocking on the foreground thread
                            # - Prevent foreground thread from being detected as the scheduler thread
                            super_resource_download = Resource.download
                            def resource_download(self: Resource, *args, **kwargs) -> 'Future[ResourceRevision]':
                                _ = super_resource_download(self, *args, **kwargs)
                                result = Future()  # type: Future[ResourceRevision]
                                result.set_result(ResourceRevision.create_from_error(self, Exception('Simulated error')))
                                return result
                            with patch.object(Resource, 'download', resource_download):
                                assert is_foreground_thread()
                                with scheduler_thread_context(enabled=False):
                                    _ = await fetch_archive_url(comic2_url)
                            
                            # Process deferred event: DRGMT.group_did_add_member
                            await step_scheduler(project)
                            
                            # Ensure newly discovered group member is queued for download
                            assertEqual(2, _group_size_in_subtitle(drgm_task))
                        finally:
                            server.close()
                        
                        # Drain tasks explicitly, to avoid subtitle-related warning
                        await step_scheduler_until_done(project)


def _group_size_in_subtitle(t: DownloadResourceGroupMembersTask) -> int:
    m = re.search(r'(\d+) item\(s\)', t.subtitle)
    assert m is not None
    return int(m.group(1))


# ------------------------------------------------------------------------------
# Test: RootTask

# (TODO: Add basic tests)


# ==============================================================================
# Test: Scheduler

@skip('covered by: test_when_scheduler_thread_event_loop_crashes_then_RT_marked_as_crashed_and_scheduler_crashed_task_appears')
async def test_when_scheduler_thread_crashes_then_scheduler_crashed_task_appears():
    pass


# ==============================================================================



from crystal.tests.util.screenshots import screenshot_if_raises_deco
from unittest import TestCase


class _DummyTestCase(TestCase):
    maxDiff = None

# Advantages of these assert methods over a bare assert statement:
# 1. A better error message is provided, including the operands that were compared
# 2. A screenshot will be taken automatically upon failure
assertEqual = screenshot_if_raises_deco(_DummyTestCase().assertEqual)
assertNotEqual = screenshot_if_raises_deco(_DummyTestCase().assertEqual)
assertIn = screenshot_if_raises_deco(_DummyTestCase().assertIn)
assertNotIn = screenshot_if_raises_deco(_DummyTestCase().assertNotIn)
assertRaises = screenshot_if_raises_deco(_DummyTestCase().assertRaises)



import sys
import urllib.parse
from urllib.parse import ParseResult, urlunparse


def patch_urlparse_to_never_raise_valueerror():
    super_urlparse = urllib.parse.urlparse  # capture

    def urlparse(url: str, scheme: str='', allow_fragments: bool=True) -> ParseResult:
        """
        Parses a URL in the same way that urllib.parse.urlparse() does,
        but always returns a ParseResult for a str input, never raising an exception.
        
        For more information about why this function exists, see discussion at:
        * https://discuss.python.org/t/urlparse-can-sometimes-raise-an-exception-should-it/44465/3
        """
        if not isinstance(url, str):
            raise TypeError()
        try:
            url_parts = super_urlparse(url, scheme, allow_fragments)
        except ValueError:
            url_parts = ParseResult(scheme='', netloc='', path=url, params='', query='', fragment='')
        
        # Check whether ParseResult can be round-tripped back to the original URL
        if scheme != '':
            # Can't check round-trip when default scheme provided
            pass
        else:
            url2 = urlunparse(url_parts)
            if url2 != url:
                if url.endswith('#'):
                    # Ignore common case where round-trip doesn't work
                    pass
                else:
                    print(
                        f'Warning: urlparse returned output that does not '
                            f'unparse to the input: {url!r} -> {url2!r}',
                        file=sys.stderr
                    )
        
        return url_parts
    urlparse._super = super_urlparse
    
    urllib.parse.urlparse = urlparse  # monkeypatch



"""
HTML parser implementation that uses BeautifulSoup.
"""

from collections.abc import Callable
from crystal.doc.css import (
    CssDocument, parse_css_and_links_from_style_attribute,
    parse_css_and_links_from_style_tag,
)
from crystal.doc.generic import Document, Link
from crystal.doc.html import HtmlParserType
from crystal.util.fastsoup import (
    BeautifulFastSoup, FastSoup, FindFunc, LxmlFastSoup, name_of_tag,
    parse_html, Tag,
)
from dataclasses import dataclass
import json
import re
from re import Match
from typing import List, Optional, Tuple, Type
from urllib.parse import urlparse

_PARSER_LIBRARY_T_CHOICES = (
    LxmlFastSoup,
    BeautifulFastSoup,
)  # type: Tuple[Type[FastSoup], ...]


_ON_CLICK_RE = re.compile(r'''(?i)([a-zA-Z.]*\.(?:href|location)) *= *(['"])([^'"]*)['"] *;?$''')
_TEXT_JAVASCRIPT_RE = re.compile(r'(?i)^text/javascript$')
_QUOTED_HTTP_LINK_RE = re.compile(
    r'(?i)(?:'
        r'(")((?:'                          # "..."
                   r'\\?/\\?/[^/"][^"]*|'       # //...
            r'https?:\\?/\\?/(?:[^/"][^"]*)?'   # https://...
        r'))"|'
        r"(')("                             # '...'
                   r"\\?/\\?/[^/'][^']*|"       # //...
            r"https?:\\?/\\?/(?:[^/'][^']*)?"   # https://...
        r")'"
    r')'
)
ABSOLUTE_HTTP_LINK_RE = re.compile(r'''(?i)^(https?://.+)$''')

PROBABLE_EMBEDDED_URL_RE = re.compile(r'(?i)\.(gif|jpe?g|png|svg|js|css)(?:\?[^/?]*)?$')


@dataclass
class _XPaths:
    STYLE_XP: FindFunc
    STYLE_EQ_STAR_XP: FindFunc
    BACKGROUND_EQ_STAR_XP: FindFunc
    SRC_EQ_STAR_XP: FindFunc
    IMG_SRCSET_EQ_STAR_XP: FindFunc
    SOURCE_SRCSET_EQ_STAR_XP: FindFunc
    HREF_EQ_STAR_XP: FindFunc
    ONCLICK_EQ_ELLIPSIS_XP: FindFunc
    SCRIPT_STRING_EQ_QUOTED_HTTP_LINK_XP: FindFunc
    STAR_XP: FindFunc

_XPS_FOR_PARSER_LIBRARY_T = {T: _XPaths(
    STYLE_XP = T.find_all_compile(
        'style'),
    STYLE_EQ_STAR_XP = T.find_all_compile(
        style=True),
    BACKGROUND_EQ_STAR_XP = T.find_all_compile(
        background=True),
    SRC_EQ_STAR_XP = T.find_all_compile(
        src=True),
    IMG_SRCSET_EQ_STAR_XP = T.find_all_compile(
        'img', srcset=True),
    SOURCE_SRCSET_EQ_STAR_XP = T.find_all_compile(
        'source', srcset=True),
    HREF_EQ_STAR_XP = T.find_all_compile(
        href=True),
    ONCLICK_EQ_ELLIPSIS_XP = T.find_all_compile(
        onclick=_ON_CLICK_RE),
    SCRIPT_STRING_EQ_QUOTED_HTTP_LINK_XP = T.find_all_compile(
        'script', string=_QUOTED_HTTP_LINK_RE),
    STAR_XP = T.find_all_compile(),
) for T in _PARSER_LIBRARY_T_CHOICES}


TEXT_LINK_TYPE_TITLE = 'Link'
FAVICON_TYPE_TITLE = 'Icon'


def parse_html_and_links(
        html_bytes: bytes, 
        declared_charset: str | None,
        parser_type: HtmlParserType,
        ) -> 'Optional[Tuple[Document, List[Link]]]':
    try:
        html = parse_html(
            html_bytes,
            from_encoding=declared_charset,
            parser_type=parser_type,
        )
    except Exception as e:
        return None
    
    XPS = _XPS_FOR_PARSER_LIBRARY_T[type(html)]
    
    links = []  # type: List[Link]
    pre_stringify_actions = []  # type: List[Callable[[], None]]
    
    # <style>...</style>
    for tag in XPS.STYLE_XP(html):
        style_body = html.tag_string(tag)
        if style_body is None:
            continue
        (css_doc, css_links) = parse_css_and_links_from_style_tag(style_body)
        links.extend(css_links)
        pre_stringify_actions.append(_update_style_tag_string_func(html, tag, css_doc))
    
    # <* style=*>
    for tag in XPS.STYLE_EQ_STAR_XP(html):
        style_value = html.tag_attrs(tag)['style']
        if not isinstance(style_value, str):
            continue
        (css_doc, css_links) = parse_css_and_links_from_style_attribute(style_value)
        links.extend(css_links)
        pre_stringify_actions.append(_update_style_attr_value_func(html, tag, css_doc))
    
    # <* background=*>
    for tag in XPS.BACKGROUND_EQ_STAR_XP(html):
        embedded = True
        title = None
        type_title = 'Background Image'
        links.append(HtmlLink.create_from_tag(tag, html, 'background', type_title, title, embedded))
    
    # <source srcset=*>, before <img src=*>
    for tag in XPS.SOURCE_SRCSET_EQ_STAR_XP(html):
        links.extend(_process_srcset_attr(html, tag))
    
    # <* src=*>
    for tag in XPS.SRC_EQ_STAR_XP(html):
        tag_name = html.tag_name(tag)  # cache
        tag_attrs = html.tag_attrs(tag)  # cache
        
        embedded = True
        if tag_name == 'img':
            title = _get_image_tag_title(html, tag)
            type_title = 'Image'
        elif tag_name == 'iframe':
            title = _assert_str(tag_attrs['name']) if 'name' in tag_attrs else None
            type_title = 'IFrame'
        elif tag_name == 'frame':
            title = _assert_str(tag_attrs['name']) if 'name' in tag_attrs else None
            type_title = 'Frame'
        elif tag_name == 'input' and 'type' in tag_attrs and tag_attrs['type'] == 'image':
            title = _get_image_tag_title(html, tag)
            type_title = 'Form Image'
        else:
            title = None
            type_title = 'Unknown Embedded (%s)' % tag_name
        if tag_name == 'script' and 'integrity' in tag_attrs:
            rewrite_side_effect = _remove_integrity_attr_func(html, tag)
        else:
            rewrite_side_effect = None
        links.append(HtmlLink.create_from_tag(
            tag, html, 'src', type_title, title, embedded,
            rewrite_side_effect=rewrite_side_effect))
    
    # <img srcset=*>
    for tag in XPS.IMG_SRCSET_EQ_STAR_XP(html):
        links.extend(_process_srcset_attr(html, tag))
    
    # <* href=*>
    for tag in XPS.HREF_EQ_STAR_XP(html):
        tag_name = html.tag_name(tag)  # cache
        tag_attrs = html.tag_attrs(tag)  # cache
        
        relative_url = _assert_str(tag_attrs['href'])
        relative_url_path = urlparse(relative_url).path
        embedded = False
        if tag_name == 'a':
            title = html.tag_string(tag)
            type_title = TEXT_LINK_TYPE_TITLE  # 'Link'
        elif tag_name == 'link' and (
                ('rel' in tag_attrs and 'stylesheet' in tag_attrs['rel']) or (
                 'type' in tag_attrs and tag_attrs['type'] == 'text/css') or (
                 relative_url_path.endswith('.css'))):
            title = None
            type_title = 'Stylesheet'
            embedded = True
        elif tag_name == 'link' and (
                    ('rel' in tag_attrs and any([x in tag_attrs['rel'] for x in (
                        'shortcut icon',
                        'icon',
                        'apple-touch-icon')])) or 
                    (relative_url_path.endswith('.ico') or 
                        relative_url_path.endswith('.png'))
                ):
            title = None
            type_title = FAVICON_TYPE_TITLE
            embedded = True
        elif tag_name == 'link' and (
                'rel' in tag_attrs and 'preload' in tag_attrs['rel']):
            title = None
            type_title = 'Preload'
            embedded = True
        elif tag_name == 'link' and 'rel' in tag_attrs:
            def format_rel(rel: str | list[str]) -> str:
                if isinstance(rel, str):
                    return rel
                elif isinstance(rel, list):
                    return ','.join(rel)
                else:
                    raise ValueError()
            
            title = None
            type_title = 'Unknown Link (rel=%s)' % format_rel(tag_attrs['rel'])
        else:
            title = None
            type_title = 'Unknown Href (%s)' % tag_name
        if tag_name == 'link' and 'integrity' in tag_attrs:
            rewrite_side_effect = _remove_integrity_attr_func(html, tag)
        else:
            rewrite_side_effect = None
        links.append(HtmlLink.create_from_tag(
            tag, html, 'href', type_title, title, embedded,
            rewrite_side_effect=rewrite_side_effect))
    
    # <* onclick='*.location = "*";'>
    # This type of link is used on:
    # - fanfiction.net
    # - http://niko-niko.net/mg/shows.html
    for tag in XPS.ONCLICK_EQ_ELLIPSIS_XP(html):
        tag_attrs = html.tag_attrs(tag)  # cache
        
        matcher = _ON_CLICK_RE.search(_assert_str(tag_attrs['onclick']))
        assert matcher is not None
        def process_match(matcher: Match) -> None:
            def replace_url_in_old_attr_value(url: str, old_attr_value: str) -> str:
                q = matcher.group(2)
                return matcher.group(1) + ' = ' + q + url + q
            
            relative_url = matcher.group(3)
            title = _assert_str(tag_attrs['value']) if 'value' in tag_attrs else None
            type_title = 'Clickable'
            embedded = False
            links.append(HtmlLink.create_from_complex_tag(
                tag, html, 'onclick', type_title, title, embedded,
                relative_url, replace_url_in_old_attr_value))
        process_match(matcher)
    
    # 1. <script [type="text/javascript"]>..."http(s)://**"...</script>
    #   - This type of link is used on: http://*.daportfolio.com/
    # 2. <script [type="text/javascript"]>..."//**"...</script>
    #   - This type of link is used on: https://blog.calm.com/take-a-deep-breath
    for tag in XPS.SCRIPT_STRING_EQ_QUOTED_HTTP_LINK_XP(html):
        tag_attrs = html.tag_attrs(tag)  # cache
        
        if 'type' in tag_attrs:
            if not _TEXT_JAVASCRIPT_RE.fullmatch(_assert_str(tag_attrs['type'])):
                continue
        
        tag_string = html.tag_string(tag)
        if tag_string is None:
            continue
        matches = _QUOTED_HTTP_LINK_RE.findall(tag_string)
        for match in matches:
            def process_str_match(match: str) -> None:
                q = match[0] or match[2]
                old_string_literal = q + (match[1] or match[3]) + q
                
                def replace_url_in_old_attr_value(url: str, old_attr_value: str) -> str:
                    quoted_url = json.dumps(url)  # type: ignore[attr-defined]
                    if q == "'":
                        new_string_literal = q + quoted_url[1:-1].replace(q, '\\' + q) + q
                    else:  # q == '"' or something else
                        new_string_literal = quoted_url
                    
                    return old_attr_value.replace(old_string_literal, new_string_literal)
                
                try:
                    relative_url = json.loads('"' + (match[1] or match[3]) + '"')  # type: ignore[attr-defined]
                except ValueError:
                    # Failed to parse JavaScript string literal
                    return
                title = None
                type_title = 'Script Reference'
                embedded = PROBABLE_EMBEDDED_URL_RE.search(relative_url) is not None
                links.append(HtmlLink.create_from_complex_tag(
                    tag, html, 'string', type_title, title, embedded,
                    relative_url, replace_url_in_old_attr_value))
            process_str_match(match)
    
    # <* *="http(s)://**">
    # This type of link is used on: https://blog.calm.com/take-a-deep-breath
    # where the attribute name is "data-url".
    seen_tags_and_attr_names = {
        (_IdentityKey(link.tag), link.attr_name) for link in links
        if isinstance(link, HtmlLink)
    }  # capture
    for tag in XPS.STAR_XP(html):
        tag_ident = _IdentityKey(tag)  # cache
        for (attr_name, attr_value) in html.tag_attrs(tag).items():
            if (tag_ident, attr_name) in seen_tags_and_attr_names:
                continue
            if not isinstance(attr_value, str):
                # HACK: BeautifulSoup has been observed to provide a 
                #       List[str] value for the "class" attribute...
                continue
            matcher = ABSOLUTE_HTTP_LINK_RE.fullmatch(attr_value)
            if not matcher:
                continue
            
            relative_url = matcher.group(1)
            title = None
            type_title = 'Attribute Reference'
            embedded = PROBABLE_EMBEDDED_URL_RE.search(relative_url) is not None
            links.append(HtmlLink.create_from_tag(
                tag, html, attr_name, type_title, title, embedded))
    
    doc = HtmlDocument(html, pre_stringify_actions=pre_stringify_actions)
    return (doc, links)


def _update_style_tag_string_func(html: FastSoup, tag: Tag, css_doc: CssDocument) -> Callable[[], None]:
    return lambda: html.set_tag_string(tag, str(css_doc))


def _update_style_attr_value_func(html: FastSoup, tag: Tag, css_doc: CssDocument) -> Callable[[], None]:
    def update_style_attr_value() -> None:
        html.tag_attrs(tag)['style'] = str(css_doc)
    return update_style_attr_value


def _remove_integrity_attr_func(html: FastSoup, tag: Tag) -> Callable[[], None]:
    def remove_integrity_attr() -> None:
        del html.tag_attrs(tag)['integrity']
    return remove_integrity_attr


def _get_image_tag_title(html: FastSoup, tag: Tag) -> str | None:
    tag_attrs = html.tag_attrs(tag)  # cache
    
    if 'alt' in tag_attrs:
        return _assert_str(tag_attrs['alt'])
    elif 'title' in tag_attrs:
        return _assert_str(tag_attrs['title'])
    else:
        return None


def _process_srcset_attr(html: FastSoup, img_or_source_tag: Tag) -> 'List[HtmlLink]':
    srcset = _parse_srcset_str(_assert_str(html.tag_attrs(img_or_source_tag)['srcset']))
    if srcset is None:
        return []
    
    links = []
    
    def process_candidate(parts: list[str]) -> None:
        nonlocal links, srcset
        
        def replace_url_in_old_attr_value(url: str, old_attr_value: str) -> str:
            nonlocal parts, srcset
            parts[0] = url  # reinterpret
            assert srcset is not None  # help mypy
            return _format_srcset_str(srcset)
        
        relative_url = parts[0]
        tag_name = name_of_tag(img_or_source_tag)  # cache
        if tag_name == 'img':
            title = _get_image_tag_title(html, img_or_source_tag)
            type_title = 'Image'
        elif tag_name == 'source':
            title = None
            type_title = 'Image Source'
        else:
            raise ValueError()
        embedded = True
        links.append(HtmlLink.create_from_complex_tag(
            img_or_source_tag, html, 'srcset', type_title, title, embedded,
            relative_url, replace_url_in_old_attr_value))
    
    candidates = srcset
    for c in candidates:
        parts = c
        process_candidate(parts)
    
    return links


# NOTE: Each URL in a srcset may itself contain a comma
#       (due to data: URLs) so cannot just split on comma
_SRCSET_PART_RE = re.compile(r' *(data:[^ ]+|[^ ,]+)(?: +([^ ,]+))? *,')

def _parse_srcset_str(srcset_str: str) -> list[list[str]] | None:
    candidates = []
    for parts in _SRCSET_PART_RE.findall(srcset_str + ','):
        if parts[1] == '':
            candidates.append([parts[0]])
        else:
            candidates.append([parts[0], parts[1]])
    return candidates


def _format_srcset_str(srcset: list[list[str]]) -> str:
    return ','.join([' '.join(parts) for parts in srcset])


class HtmlDocument(Document):
    def __init__(self,
            html: FastSoup,
            *, is_html: bool=True,
            pre_stringify_actions: list[Callable[[], None]] | None=None,
            ) -> None:
        self._html = html
        self._is_html = is_html
        self._pre_stringify_actions = pre_stringify_actions or []
    
    def try_insert_script(self, script_url: str) -> bool:
        def create_script(html: FastSoup) -> Tag:
            script = html.new_tag('script')
            html.tag_attrs(script)['src'] = script_url
            return script
        return self._try_insert_html_element(create_script) is not None
    
    def try_insert_favicon_link(self, favicon_url: str) -> Link | None:
        def create_link(html: FastSoup) -> Tag:
            link = self._html.new_tag('link')
            self._html.tag_attrs(link)['rel'] = 'icon'
            self._html.tag_attrs(link)['href'] = favicon_url
            return link
        link_tag = self._try_insert_html_element(create_link)
        if link_tag is None:
            return None
        link = HtmlLink.create_from_tag(
            link_tag, self._html, 'href', FAVICON_TYPE_TITLE, None, True)
        return link
    
    def _try_insert_html_element(self, create_element_func: Callable[[FastSoup], Tag]) -> Tag | None:
        if not self._is_html:
            # Don't try to insert an HTML link into a non-HTML document,
            # such as an XML document
            return False
        
        first_element = self._html.find(True)
        if first_element is not None:
            new_element = create_element_func(self._html)
            self._html.tag_insert_before(first_element, new_element)
            return new_element
        else:
            return None
    
    def __str__(self) -> str:
        for action in self._pre_stringify_actions:
            action()
        return str(self._html)


# TODO: Split this internally into three subclasses
class HtmlLink(Link):
    """
    Represents a link in a (usually-HTML) resource.
    """
    @staticmethod
    def create_from_tag(
            tag: Tag,
            tag_doc: FastSoup,
            attr_name: str,
            type_title: str,
            title: str | None,
            embedded: bool,
            *, rewrite_side_effect: Callable[[], None] | None=None,
            ) -> 'HtmlLink':
        """
        Creates a link that is derived from the attribute of an HTML element.
        
        Arguments:
        * relative_url - URL or URI referenced by this link, often relative.
        * type_title - displayed title for this link's type.
        * title - displayed title for this link, or None.
        * embedded - whether this link refers to an embedded resource.
        """
        if (tag is None or attr_name is None or type_title is None or
                embedded not in (True, False)):
            raise ValueError()
        return HtmlLink(
            None, tag, tag_doc, attr_name, type_title, title, embedded,
            rewrite_side_effect=rewrite_side_effect)
    
    @staticmethod
    def create_from_complex_tag(
            tag: Tag,
            tag_doc: FastSoup,
            attr_name: str,
            type_title: str,
            title: str | None,
            embedded: bool,
            relative_url: str,
            replace_url_in_old_attr_value: Callable[[str, str], str]
            ) -> 'HtmlLink':
        """
        See HtmlLink.create_from_tag()
        
        Extra Arguments:
        * replace_url_in_old_attr_value --
            function that takes a URL and returns the appropriate
            value for the underlying tag's attribute.
        """
        if (tag is None or attr_name is None or not callable(replace_url_in_old_attr_value) or 
                type_title is None or embedded not in (True, False)):
            raise ValueError()
        return HtmlLink(
            relative_url, tag, tag_doc, attr_name, type_title, title, embedded,
            replace_url_in_old_attr_value=replace_url_in_old_attr_value)
    
    @staticmethod
    def create_external(
            relative_url: str,
            type_title: str,
            title: str | None,
            embedded: bool
            ) -> 'HtmlLink':
        """
        Creates a external link that is not reflected in the original HTML content.
        
        Arguments:
        * relative_url - URL or URI referenced by this link, often relative.
        * type_title - displayed title for this link's type.
        * title - displayed title for this link, or None.
        * embedded - whether this link refers to an embedded resource.
        """
        if relative_url is None or type_title is None or embedded not in (True, False):
            raise ValueError
        return HtmlLink(relative_url, None, None, None, type_title, title, embedded)
    
    def __init__(self,
            relative_url: str | None,
            tag: Tag | None,
            tag_doc: FastSoup | None,
            attr_name: str | None,
            type_title: str,
            title: str | None,
            embedded: bool,
            *, replace_url_in_old_attr_value: Callable[[str, str], str] | None=None,
            rewrite_side_effect: Callable[[], None] | None=None,
            ) -> None:
        self._relative_url = relative_url
        self._tag = tag
        self._tag_doc = tag_doc
        self._attr_name = attr_name
        self._replace_url_in_old_attr_value = replace_url_in_old_attr_value
        self._rewrite_side_effect = rewrite_side_effect
        
        self.title = title
        self.type_title = type_title
        self.embedded = embedded
    
    def _get_relative_url(self) -> str:
        if self._relative_url:
            return self._relative_url
        else:
            # NOTE: Immediately strip whitespace from relative URLs inside HTML tags,
            #       before they are resolved against any base URL. Both preceding
            #       and trailing whitespace has been observed in the wild.
            return self._attr_value.strip()
    def _set_relative_url(self, url: str) -> None:
        if self._relative_url and not self._replace_url_in_old_attr_value:
            self._relative_url = url
        else:
            if self._replace_url_in_old_attr_value:
                attr_value = self._replace_url_in_old_attr_value(url, self._attr_value)
            else:
                attr_value = url
            self._attr_value = attr_value
            self._relative_url = url
        if self._rewrite_side_effect is not None:
            self._rewrite_side_effect()
    relative_url = property(_get_relative_url, _set_relative_url)
    
    @property
    def tag(self) -> Tag | None:
        return self._tag
    
    @property
    def attr_name(self) -> str | None:
        return self._attr_name
    
    def _get_attr_value(self) -> str:
        assert self._tag is not None
        assert self._tag_doc is not None
        assert self._attr_name is not None
        if self._attr_name == 'string':
            tag_string = self._tag_doc.tag_string(self._tag)
            assert tag_string is not None
            return tag_string
        else:
            tag_attrs = self._tag_doc.tag_attrs(self._tag)
            return _assert_str(tag_attrs[self._attr_name])
    def _set_attr_value(self, attr_value: str) -> None:
        assert self._tag is not None
        assert self._tag_doc is not None
        assert self._attr_name is not None
        if self._attr_name == 'string':
            self._tag_doc.set_tag_string(self._tag, attr_value)
        else:
            tag_attrs = self._tag_doc.tag_attrs(self._tag)
            tag_attrs[self._attr_name] = attr_value
    _attr_value = property(_get_attr_value, _set_attr_value)
    
    def __repr__(self) -> str:
        # TODO: Update repr to include new constructor parameters
        return 'HtmlLink({},{},{},{})'.format(repr(self.relative_url), repr(self.type_title), repr(self.title), repr(self.embedded))


class _IdentityKey:
    """
    Wraps an object, so that equality comparisons and hash operations
    look at the object's identity only.
    """
    # Optimize per-instance memory use, since there may be many _IdentityKey objects
    __slots__ = ('_obj',)
    
    def __init__(self, obj: object) -> None:
        self._obj = obj
    
    def __hash__(self) -> int:
        return id(self._obj)
    
    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, _IdentityKey) and
            self._obj is other._obj
        )


def _assert_str(value: object) -> str:
    assert isinstance(value, str)
    return value



from crystal.util.xos import is_linux
import subprocess


def set(itempath: str, attr_name: str, attr_value: str) -> None:
    """
    Raises:
    * GioNotAvailable
    * UnrecognizedGioAttributeError
    """
    assert is_linux()
    try:
        subprocess.run(
            ['gio', 'set', itempath, attr_name, attr_value],
            check=True,
            capture_output=True,
        )
    except FileNotFoundError:
        raise GioNotAvailable()
    except subprocess.CalledProcessError as e:
        if (e.stderr.startswith(b'gio: Setting attribute ') and
                e.stderr.endswith(b'not supported\n')):
            raise UnrecognizedGioAttributeError()
        else:
            raise


class GioNotAvailable(Exception):
    pass


class UnrecognizedGioAttributeError(Exception):
    pass



import re

_URL_WITH_REPLYTOCOM = re.compile(r'^(https?://[^?]+)\?replytocom=[0-9]+/?$')

_URL_WITH_POSTTYPE_SLASH = re.compile(r'^(https?://[^?]+\?post_type=[a-zA-Z_]*)((?:/|%2F)*)$')

def normalize_url(old_url: str, **kwargs) -> str:
    # Normalize "reply to comment" links to strip out parameter,
    # which won't be used anyway if page is loaded with JavaScript
    # 
    # Read more about this parameter at:
    # https://www.namehero.com/blog/filtering-out-replytocom-bots-on-wordpress/
    # 
    # https?://__DOMAIN__/...?replytocom=#
    m = _URL_WITH_REPLYTOCOM.fullmatch(old_url)
    if m is not None:
        # Chop off replytocom=... part
        return m.group(1)
    
    # NextGEN Gallery (nggallery):
    # Prevent percent-encode of slash within ?post_type=... segment
    m = _URL_WITH_POSTTYPE_SLASH.fullmatch(old_url)
    if m is not None:
        return m.group(1) + m.group(2).replace('%2F', '/')
    
    return old_url



import math
import os
import platform
import subprocess
from typing import Optional

# === OS Detection ===

def is_mac_os() -> bool:
    return (platform.system() == 'Darwin')


def is_windows() -> bool:
    return (platform.system() == 'Windows')


def is_linux() -> bool:
    return (platform.system() == 'Linux')


def windows_major_version() -> int | None:
    """
    Returns the major version number of Windows, or None if unknown.
    
    Examples:
    * 7 -- Windows 7
    * 8 -- Windows 8, Windows 8.1
    * 10 -- Windows 10
    * None -- unknown Windows
    """
    if not is_windows():
        return None
    
    release_str = platform.release()
    try:
        release_number = float(release_str)
    except ValueError:
        return None
    else:
        return math.floor(release_number)


def mac_version() -> list[int] | None:
    if not is_mac_os():
        return None
    
    return [int(x) for x in platform.mac_ver()[0].split('.')]


# === Linux Distro Detection ===

_is_gnome = None  # type: Optional[bool]

def is_gnome() -> bool:
    global _is_gnome
    if not is_linux():
        return False
    if _is_gnome is None:
        try:
            subprocess.run(
                ['gnome-shell', '--version'],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=True)
        except FileNotFoundError:
            _is_gnome = False
        else:
            _is_gnome = True
    return _is_gnome


def is_kde_or_non_gnome() -> bool:
    if not is_linux():
        return False
    return not is_gnome()


# === wxPython Backend Detection ===

def is_wx_gtk() -> bool:
    return is_linux()


# === Continuous Integration Detection ===

def is_ci() -> bool:
    return os.environ.get('CI') == 'true'


def is_asan() -> bool:
    return os.environ.get('CRYSTAL_ADDRESS_SANITIZER') == 'True'



from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager, contextmanager
from crystal.browser.new_root_url import fields_hide_hint_when_focused
from crystal.model import Project, Resource, RootResource
from crystal.task import DownloadResourceGroupTask
from crystal.tests.util.asserts import *
from crystal.tests.util.controls import click_button, click_checkbox, TreeItem
from crystal.tests.util.server import MockHttpServer, served_project
from crystal.tests.util.subtests import (
    awith_subtests, SubtestsContext, with_subtests,
)
from crystal.tests.util.tasks import (
    append_deferred_top_level_tasks, clear_top_level_tasks_on_exit,
    scheduler_disabled, wait_for_download_to_start_and_finish,
)
from crystal.tests.util.wait import (
    DEFAULT_WAIT_PERIOD, first_child_of_tree_item_is_not_loading_condition,
    tree_has_no_children_condition, wait_for,
)
from crystal.tests.util.windows import (
    EntityTree, NewRootUrlDialog, OpenOrCreateDialog,
)
import crystal.url_input
from crystal.url_input import _candidate_urls_from_user_input as EXPAND
from crystal.util.wx_dialog import mocked_show_modal
from crystal.util.wx_window import SetFocus
import os
import time
from typing import Dict, Optional, Self, Union
from unittest import skip
from unittest.mock import ANY, patch
from urllib.parse import urlparse
import wx

# === Test: Create & Delete Standalone ===

async def test_can_create_root_url(
        *, ensure_revisions_not_deleted: bool=False,
        add_surrounding_whitespace: bool=False) -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Create root URL
            if True:
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                () = root_ti.Children
                
                assert mw.new_root_url_button.Enabled
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                nud.do_not_download_immediately()
                
                # Ensure prepopulates reasonable information
                assert '' == nud.url_field.Value
                assert '' == nud.name_field.Value
                #assert None == ngd.source
                assert nud.url_field.HasFocus  # default focused field
                
                SetFocus(nud.name_field, None)
                nud.name_field.Value = 'Home'
                SetFocus(nud.url_field, nud.name_field)
                nud.url_field.Value = (
                    (' ' if add_surrounding_whitespace else '') +
                    home_url +
                    (' ' if add_surrounding_whitespace else '')
                )
                await nud.ok()
                
                # Ensure appearance is correct
                home_ti = root_ti.find_child(home_url, project.default_url_prefix)
                assert f'{urlparse(home_url).path} - Home' == home_ti.Text
                await _assert_tree_item_icon_tooltip_contains(home_ti, 'root URL')
                await _assert_tree_item_icon_tooltip_contains(home_ti, 'Undownloaded')
                assert f'URL: {home_url}' in (home_ti.Tooltip('label') or '')
                
                # Ensure new root URL is selected automatically,
                # given that nothing was previously selected
                selected_ti = TreeItem.GetSelection(mw.entity_tree.window)
                assert selected_ti == home_ti
            
            if ensure_revisions_not_deleted:
                # Download a revision of the root URL
                home_ti.SelectItem()
                await mw.click_download_button()
                await wait_for_download_to_start_and_finish(mw.task_tree)
                await _assert_tree_item_icon_tooltip_contains(home_ti, 'Fresh')
            
            # Forget root URL
            if True:
                home_ti.SelectItem()
                assert mw.forget_button.IsEnabled()
                click_button(mw.forget_button)
                
                # Ensure cannot find root URL
                assert None == root_ti.try_find_child(home_url, project.default_url_prefix)
                selected_ti = TreeItem.GetSelection(mw.entity_tree.window)
                assert (selected_ti is None) or (selected_ti == root_ti)
            
            if ensure_revisions_not_deleted:
                # Recreate the root URL
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                nud.do_not_download_immediately()
                nud.name_field.Value = 'Home'
                nud.url_field.Value = home_url
                await nud.ok()
                
                # 1. Ensure appearance is correct
                # 2. Ensure previously downloaded revisions still exist
                home_ti = root_ti.find_child(home_url, project.default_url_prefix)
                assert f'{urlparse(home_url).path} - Home' == home_ti.Text
                await _assert_tree_item_icon_tooltip_contains(home_ti, 'root URL')
                await _assert_tree_item_icon_tooltip_contains(home_ti, 'Fresh')
                assert f'URL: {home_url}' in (home_ti.Tooltip('label') or '')


@skip('covered by: test_can_create_root_url')
async def test_can_forget_root_url() -> None:
    pass


async def test_when_forget_root_url_then_revisions_of_that_url_are_not_deleted() -> None:
    await test_can_create_root_url(ensure_revisions_not_deleted=True)


async def test_given_url_with_surrounding_whitespace_when_create_root_url_then_surrounding_whitespace_ignored() -> None:
    await test_can_create_root_url(add_surrounding_whitespace=True)


# === Test: Create & Delete from Links ===

async def test_given_resource_node_with_links_can_create_new_root_url_to_label_link() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Create home URL
            if True:
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                () = root_ti.Children
                
                assert mw.new_root_url_button.Enabled
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                nud.do_not_download_immediately()
                
                nud.name_field.Value = 'Home'
                nud.url_field.Value = home_url
                await nud.ok()
                (home_ti,) = root_ti.Children
            
            # Expand home URL
            home_ti.Expand()
            await wait_for_download_to_start_and_finish(mw.task_tree)
            assert first_child_of_tree_item_is_not_loading_condition(home_ti)()
            
            # Select the Atom Feed link from the home URL
            atom_feed_ti = home_ti.find_child(atom_feed_url, project.default_url_prefix)  # ensure did find sub-resource
            atom_feed_ti.SelectItem()
            
            # Create a root resource to label the link
            if True:
                assert mw.new_root_url_button.Enabled
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                nud.do_not_download_immediately()
                
                # Ensure prepopulates reasonable information
                assert atom_feed_url == nud.url_field.Value  # default pattern = (from resource)
                assert 'Feed' == nud.name_field.Value  # default name = (from first text link)
                #assert 'Home' == nud.source  # default source = (from resource parent)
                assert nud.url_field.HasFocus  # default focused field
                
                # Input new name
                nud.name_field.Value = 'Atom Feed'
                
                await nud.ok()
            
            # 1. Ensure the new root resource does now label the link
            # 2. Ensure the labeled link is selected immediately after closing the add URL dialog
            atom_feed_ti = home_ti.find_child(atom_feed_url, project.default_url_prefix)  # ensure did find sub-resource
            assert (
                # title format of labeled sub-resource
                f'{urlparse(atom_feed_url).path} - Atom Feed' ==
                atom_feed_ti.Text)
            assert atom_feed_ti.IsSelected()
            
            # Forget the root resource to unlabel the link
            if True:
                assert atom_feed_ti.IsSelected()
                assert mw.forget_button.IsEnabled()
                click_button(mw.forget_button)
                
                # 1. Ensure can find the unlabeled link
                # 2. Ensure that unlabeled link is selected immediately after forgetting the root resource
                atom_feed_ti = home_ti.find_child(atom_feed_url, project.default_url_prefix)  # ensure did find sub-resource
                assert (
                    # title format of unlabeled sub-resource
                    f'{urlparse(atom_feed_url).path} - Unknown Link (rel=alternate), Link: Feed, Link: Atom Feed' ==
                    atom_feed_ti.Text)
                assert atom_feed_ti.IsSelected()


@skip('covered by: test_given_resource_node_with_links_can_create_new_root_url_to_label_link')
async def test_given_resource_node_with_link_labeled_as_root_url_can_easily_forget_the_root_url_to_unlabel_the_link() -> None:
    pass


# === Test: New URL Options ===

async def test_when_add_url_then_downloads_url_immediately_by_default() -> None:
    with _served_simple_site_with_2_urls() as (home_url, _):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            
            assert nud.download_immediately_checkbox.Value
            
            nud.url_field.Value = home_url
            await nud.ok()
            
            # Ensure started downloading
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            home_ti = root_ti.find_child(home_url, project.default_url_prefix)
            await wait_for_download_to_start_and_finish(mw.task_tree)
            await _assert_tree_item_icon_tooltip_contains(home_ti, 'Fresh')


async def test_when_add_url_then_can_avoid_downloading_url_with_1_extra_click() -> None:
    with _served_simple_site_with_2_urls() as (home_url, _):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            
            nud.url_field.Value = home_url
            click_checkbox(nud.download_immediately_checkbox)  # extra click #1
            assert not nud.download_immediately_checkbox.Value
            await nud.ok()
            
            # Ensure did NOT start downloading
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            home_ti = root_ti.find_child(home_url, project.default_url_prefix)
            assert tree_has_no_children_condition(mw.task_tree)()
            await _assert_tree_item_icon_tooltip_contains(home_ti, 'Undownloaded')


async def test_when_add_url_at_site_root_then_can_download_site_with_1_extra_click() -> None:
    with _served_simple_site_with_2_urls() as (home_url, _), scheduler_disabled():
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                assert mw.new_root_url_button.Enabled
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                
                nud.url_field.Value = home_url  # at site root
                assert nud.download_immediately_checkbox.Value
                assert nud.create_group_checkbox.Enabled
                click_checkbox(nud.create_group_checkbox)  # extra click #1
                assert nud.create_group_checkbox.Value
                await nud.ok()
                append_deferred_top_level_tasks(project)
                
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                home_ti = root_ti.find_child(home_url, project.default_url_prefix)
                site_g = root_ti.find_child(home_url + '**', project.default_url_prefix)
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                # (Do NOT wait for group to finish downloading)


async def test_when_add_url_not_at_site_root_then_cannot_download_site_or_create_group_for_site() -> None:
    with _served_simple_site_with_2_urls() as (home_url, image_url):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            
            nud.url_field.Value = image_url  # NOT at site root
            assert not nud.create_group_checkbox.Enabled
            await nud.cancel()


async def test_when_add_url_at_site_root_then_can_create_group_for_site_but_not_download_it_with_extra_clicks() -> None:
    with _served_simple_site_with_2_urls() as (home_url, _), scheduler_disabled():
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            with clear_top_level_tasks_on_exit(project):
                assert mw.new_root_url_button.Enabled
                click_button(mw.new_root_url_button)
                nud = await NewRootUrlDialog.wait_for()
                
                nud.url_field.Value = home_url  # at site root
                assert nud.create_group_checkbox.Enabled
                click_checkbox(nud.create_group_checkbox)  # extra click #1
                assert nud.create_group_checkbox.Value
                click_checkbox(nud.download_immediately_checkbox)  # extra click #2
                assert not nud.download_immediately_checkbox.Value
                await nud.ok()
                append_deferred_top_level_tasks(project)
                
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                home_ti = root_ti.find_child(home_url, project.default_url_prefix)
                site_g = root_ti.find_child(home_url + '**', project.default_url_prefix)
                () = project.root_task.children


async def test_when_edit_url_then_new_url_options_not_shown() -> None:
    with _served_simple_site_with_2_urls() as (home_url, _):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            rr = RootResource(project, 'Home', Resource(project, home_url))
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            home_ti = root_ti.find_child(home_url, project.default_url_prefix)
            home_ti.SelectItem()
            
            assert mw.edit_button.Enabled
            click_button(mw.edit_button)
            nud = await NewRootUrlDialog.wait_for()
            
            assert not nud.new_options_shown
            await nud.cancel()


@contextmanager
def _served_simple_site_with_2_urls() -> Iterator[tuple[str, str]]:
    server = MockHttpServer({
        '/': dict(
            status_code=200,
            headers=[('Content-Type', 'text/html')],
            content=b'<img src="/assets/image.png" />'
        ),
        '/assets/image.png': dict(
            status_code=200,
            headers=[('Content-Type', 'image/png')],
            content=b''
        )
    })
    with server:
        home_url = server.get_url('/')
        image_url = server.get_url('/assets/image.png')
        
        yield (home_url, image_url)


# === Test: Default URL Prefix: Load/Save ===

async def test_when_new_url_and_save_given_project_prefix_is_unset_then_sets_prefix_to_domain() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        # Case 1: HTTP URL
        assert None == project.default_url_prefix
        if True:
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            nud.url_field.Value = 'https://xkcd.com/'
            await nud.ok()
        assert 'https://xkcd.com' == project.default_url_prefix
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rrn = root_ti.find_child('https://xkcd.com/', project.default_url_prefix)
        await mw.entity_tree.clear_default_domain_from_entity_at_tree_item(rrn)
        
        # Case 2: Non-HTTP URL
        assert None == project.default_url_prefix
        if True:
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            nud.url_field.Value = 'mailto:me@example.com'
            await nud.ok()
        assert None == project.default_url_prefix


async def test_when_new_url_and_save_given_project_prefix_is_set_then_maintains_prefix() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        rr1 = RootResource(project, '', Resource(project, 'https://neocities.org/'))
        rr2 = RootResource(project, '', Resource(project, 'https://neocities.org/~distantskies/'))
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rrn1 = root_ti.find_child(rr1.resource.url, project.default_url_prefix)
        rrn2 = root_ti.find_child(rr2.resource.url, project.default_url_prefix)
        
        await mw.entity_tree.set_default_domain_to_entity_at_tree_item(rrn1)
        
        assert 'https://neocities.org' == project.default_url_prefix
        if True:
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            nud.url_field.Value = 'https://xkcd.com/'
            await nud.ok()
        assert 'https://neocities.org' == project.default_url_prefix


async def test_when_edit_url_and_save_then_maintains_prefix() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        rr1 = RootResource(project, '', Resource(project, 'https://neocities.org/'))
        rr2 = RootResource(project, '', Resource(project, 'https://neocities.org/~distantskies/'))
        rr3 = RootResource(project, '', Resource(project, 'https://xkcd.com/'))
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rrn1 = root_ti.find_child(rr1.resource.url, project.default_url_prefix)
        rrn2 = root_ti.find_child(rr2.resource.url, project.default_url_prefix)
        rrn3 = root_ti.find_child(rr3.resource.url, project.default_url_prefix)
        
        assert None == project.default_url_prefix
        for rrn in [rrn1, rrn2, rrn3]:
            if True:
                rrn.SelectItem()
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                nud = await NewRootUrlDialog.wait_for()
                
                await nud.ok()
            assert None == project.default_url_prefix
        
        await mw.entity_tree.set_default_domain_to_entity_at_tree_item(rrn1)
        assert 'https://neocities.org' == project.default_url_prefix
        for rrn in [rrn1, rrn2, rrn3]:
            if True:
                rrn.SelectItem()
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                nud = await NewRootUrlDialog.wait_for()
                
                await nud.ok()
            assert 'https://neocities.org' == project.default_url_prefix
        
        await mw.entity_tree.set_default_directory_to_entity_at_tree_item(rrn2)
        assert 'https://neocities.org/~distantskies' == project.default_url_prefix
        for rrn in [rrn1, rrn2, rrn3]:
            if True:
                rrn.SelectItem()
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                nud = await NewRootUrlDialog.wait_for()
                
                await nud.ok()
            assert 'https://neocities.org/~distantskies' == project.default_url_prefix


async def test_when_new_url_and_set_prefix_to_x_and_save_then_sets_prefix_to_x() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        
        # Case 1.1: given_project_prefix_is_unset, set_prefix_to_domain
        assert None == project.default_url_prefix
        if True:
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            nud.url_field.Value = 'https://neocities.org/'
            
            click_button(nud.options_button)
            assert (True, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert 'https://neocities.org' == project.default_url_prefix
        
        rrn = root_ti.find_child('https://neocities.org/', project.default_url_prefix)
        await mw.entity_tree.clear_default_domain_from_entity_at_tree_item(rrn)
        rrn.SelectItem()
        click_button(mw.forget_button)
        
        # Case 1.2: given_project_prefix_is_unset, set_prefix_to_directory
        assert None == project.default_url_prefix
        if True:
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            nud.url_field.Value = 'https://neocities.org/~distantskies/'
            
            click_button(nud.options_button)
            assert (True, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            click_checkbox(nud.set_as_default_directory_checkbox)
            assert (False, True) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert 'https://neocities.org/~distantskies' == project.default_url_prefix
        
        rrn = root_ti.find_child('https://neocities.org/~distantskies/', project.default_url_prefix)
        await mw.entity_tree.clear_default_directory_from_entity_at_tree_item(rrn)
        rrn.SelectItem()
        click_button(mw.forget_button)
        
        project.default_url_prefix = 'https://xkcd.com'
        
        # Case 2.1: given_project_prefix_is_set, set_prefix_to_domain
        assert 'https://xkcd.com' == project.default_url_prefix
        if True:
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            nud.url_field.Value = 'https://neocities.org/'
            
            click_button(nud.options_button)
            assert (False, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            click_checkbox(nud.set_as_default_domain_checkbox)
            assert (True, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert 'https://neocities.org' == project.default_url_prefix
        
        # Case 2.2: given_project_prefix_is_set, set_prefix_to_directory
        assert 'https://neocities.org' == project.default_url_prefix
        if True:
            assert mw.new_root_url_button.Enabled
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            nud.url_field.Value = 'https://neocities.org/~distantskies/'
            
            click_button(nud.options_button)
            assert (False, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            click_checkbox(nud.set_as_default_directory_checkbox)
            assert (False, True) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert 'https://neocities.org/~distantskies' == project.default_url_prefix


async def test_when_edit_url_and_set_prefix_to_x_and_save_then_sets_prefix_to_x() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        rr = RootResource(project, '', Resource(project, 'https://neocities.org/~distantskies/'))
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rrn = root_ti.find_child(rr.resource.url, project.default_url_prefix)
        
        # Case 1: set_prefix_to_domain
        rrn.SelectItem()
        if True:
            assert mw.edit_button.Enabled
            click_button(mw.edit_button)
            nud = await NewRootUrlDialog.wait_for()
            
            click_button(nud.options_button)
            assert (False, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            click_checkbox(nud.set_as_default_domain_checkbox)
            assert (True, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert 'https://neocities.org' == project.default_url_prefix
        
        # Case 2: set_prefix_to_directory
        rrn.SelectItem()
        if True:
            assert mw.edit_button.Enabled
            click_button(mw.edit_button)
            nud = await NewRootUrlDialog.wait_for()
            
            click_button(nud.options_button)
            assert (True, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            click_checkbox(nud.set_as_default_directory_checkbox)
            assert (False, True) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert 'https://neocities.org/~distantskies' == project.default_url_prefix
        
        # Case 3: unset_prefix
        rrn.SelectItem()
        if True:
            assert mw.edit_button.Enabled
            click_button(mw.edit_button)
            nud = await NewRootUrlDialog.wait_for()
            
            click_button(nud.options_button)
            assert (False, True) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            click_checkbox(nud.set_as_default_directory_checkbox)
            assert (False, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert None == project.default_url_prefix
        
        # Case 4: Non-HTTP URL
        rr2 = RootResource(project, '', Resource(project, 'mailto:me@example.com'))
        rrn2 = root_ti.find_child(rr2.resource.url, project.default_url_prefix)
        rrn2.SelectItem()
        if True:
            assert mw.edit_button.Enabled
            click_button(mw.edit_button)
            nud = await NewRootUrlDialog.wait_for()
            
            click_button(nud.options_button)
            assert (False, False) == (
                nud.set_as_default_domain_checkbox.Enabled,
                nud.set_as_default_directory_checkbox.Enabled
            )
            assert (False, False) == (
                nud.set_as_default_domain_checkbox.Value,
                nud.set_as_default_directory_checkbox.Value
            )
            await nud.ok()
        assert None == project.default_url_prefix


# === Test: URL Input -> Candidate URLs ===

def test_given_empty_string_then_returns_empty_string() -> None:
    with _EXPAND_enabled():
        assertEqual([''], EXPAND(''))


@with_subtests
def test_given_any_valid_url_then_returns_that_url_as_only_candidate(subtests: SubtestsContext) -> None:
    VALID_URLS = [
        'https://xkcd.com/',
        'https://xkcd.com/1/',
        'https://www.apple.com/',
        'https://www.apple.com/mac/',
        'http://localhost:8000/',
        'http://localhost:8000/logout',
        'ftp://ftp.apple.com/',
        'mailto:me@example.com',
    ]
    with _EXPAND_enabled():
        for url in VALID_URLS:
            with subtests.test(url=url):
                assertEqual([url], EXPAND(url))


@with_subtests
def test_given_a_url_that_would_be_valid_with_appended_slash_then_returns_that_modified_url_as_only_candidate(subtests: SubtestsContext) -> None:
    VALID_WITH_SLASH_URLS = [
        'https://xkcd.com',
        'https://www.apple.com',
        'http://localhost:8000',
    ]
    with _EXPAND_enabled():
        for url in VALID_WITH_SLASH_URLS:
            with subtests.test(url=url):
                assertEqual([url + '/'], EXPAND(url))


@with_subtests
def test_given_schemaless_url_without_www_prefix_then_returns_that_url_plus_that_url_with_www_as_candidates(subtests: SubtestsContext) -> None:
    CASES = [
        ('xkcd.com/', ['https://xkcd.com/', 'https://www.xkcd.com/']),
        ('apple.com/', ['https://apple.com/', 'https://www.apple.com/']),
    ]
    with _EXPAND_enabled():
        for (input, output) in CASES:
            with subtests.test(input=input):
                _assert_contains_sublist(EXPAND(input), output)


@with_subtests
def test_given_schemaless_url_with_www_prefix_then_returns_that_url_plus_that_url_without_www_as_candidates(subtests: SubtestsContext) -> None:
    CASES = [
        ('www.xkcd.com/', ['https://www.xkcd.com/', 'https://xkcd.com/']),
        ('www.apple.com/', ['https://www.apple.com/', 'https://apple.com/']),
    ]
    with _EXPAND_enabled():
        for (input, output) in CASES:
            with subtests.test(input=input):
                _assert_contains_sublist(EXPAND(input), output)


@with_subtests
def test_given_schemaless_url_then_returns_https_and_http_candidate_variations_of_url(subtests: SubtestsContext) -> None:
    CASES = [
        ('xkcd.com/', ['https://xkcd.com/', 'http://xkcd.com/']),
        ('www.apple.com/', ['https://www.apple.com/', 'http://www.apple.com/']),
        
        ('xkcd.com/1/', ['https://xkcd.com/1/', 'http://xkcd.com/1/']),
        ('www.apple.com/mac/', ['https://www.apple.com/mac/', 'http://www.apple.com/mac/']),
        
        ('xkcd.com', ['https://xkcd.com/', 'http://xkcd.com/']),
        ('www.apple.com', ['https://www.apple.com/', 'http://www.apple.com/']),
    ]
    with _EXPAND_enabled():
        for (input, output) in CASES:
            with subtests.test(input=input):
                _assert_contains_sublist(EXPAND(input), output)


def test_given_schemaless_url_without_www_prefix_then_returns_ellipsis() -> None:
    (input, output) = ('xkcd.com', ['https://xkcd.com/', 'https://www.xkcd.com/', 'http://xkcd.com/', 'http://www.xkcd.com/'])
    with _EXPAND_enabled():
        _assert_contains_sublist(EXPAND(input), output)


def test_given_schemaless_url_with_www_prefix_then_returns_ellipsis() -> None:
    (input, output) = ('www.apple.com', ['https://www.apple.com/', 'https://apple.com/', 'http://www.apple.com/', 'http://apple.com/'])
    with _EXPAND_enabled():
        _assert_contains_sublist(EXPAND(input), output)


# === Test: Validate URL upon Blur ===

async def test_given_url_input_is_empty_and_focused_when_tab_pressed_then_url_input_unfocused_and_url_input_empty_and_no_spinner_visible() -> None:
    async with _new_root_url_dialog_open() as (nud, project):
        if fields_hide_hint_when_focused():
            SetFocus(nud.url_field, None)
        else:
            assertEqual(True, nud.url_field_focused)
        assertEqual(False, nud.url_cleaner_spinner.IsShown())
        assertEqual('', nud.url_field.Value)
        
        SetFocus(nud.name_field, nud.url_field)  # simulate press tab
        assertEqual(False, nud.url_field_focused)
        assertEqual(False, nud.url_cleaner_spinner.IsShown())
        assertEqual('', nud.url_field.Value)


@awith_subtests
async def test_given_url_input_is_nonempty_and_focused_when_tab_pressed_then_url_input_unfocused_and_spinner_appears(subtests: SubtestsContext) -> None:
    URLS = [
        'xkcd.com',
        'www.apple.com',
    ]
    async with _new_root_url_dialog_open() as (nud, project):
        with _urlopen_responding_with(_UrlOpenHttpResponse(code=404, url=ANY)):
            for url in URLS:
                with subtests.test(url=url):
                    if fields_hide_hint_when_focused():
                        SetFocus(nud.url_field, None)
                    nud.url_field.Value = url
                    
                    SetFocus(nud.name_field, nud.url_field)  # simulate press tab
                    assertEqual(False, nud.url_field_focused)
                    assertEqual(True, nud.url_cleaner_spinner.IsShown())
                    
                    SetFocus(nud.url_field, nud.name_field)  # simulate press shift-tab
                    assertEqual(True, nud.url_field_focused)
                    await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)


# === Test: Resolve URL ===

@awith_subtests
async def test_given_url_input_is_nonempty_and_did_press_tab_and_spinner_is_visible_when_url_responds_with_http_200_then_spinner_disappears(subtests: SubtestsContext) -> None:
    CASES = [
        ('xkcd.com', 'https://xkcd.com/'),
        ('www.apple.com', 'https://www.apple.com/'),
    ]
    async with _new_root_url_dialog_open() as (nud, project):
        last_focused = None  # type: Optional[wx.Window]
        for (url, normalized_url) in CASES:
            with subtests.test(url=url):
                # NOTE: Fail if requests any URL beyond the original one
                with _urlopen_responding_with({url: _UrlOpenHttpResponse(code=200, url=url)}):
                    last_focused = SetFocus(nud.url_field, last_focused)
                    nud.url_field.Value = url
                    
                    last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                    assertEqual(False, nud.url_field_focused)
                    assertEqual(True, nud.url_cleaner_spinner.IsShown())
                    await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
                    assertEqual(normalized_url, nud.url_field.Value)


@awith_subtests
async def test_given_url_input_is_nonempty_without_www_and_did_press_tab_and_spinner_is_visible_when_url_responds_with_http_3xx_to_url_with_www_and_url_with_www_url_responds_with_http_200_then_url_input_replaced_with_url_with_www_and_spinner_disappears(subtests: SubtestsContext) -> None:
    CASES = [
        ('apple.com/mac/', 'https://apple.com/mac/', 'https://www.apple.com/mac/'),
        ('apple.com/', 'https://apple.com/', 'https://www.apple.com/'),
        ('apple.com', 'https://apple.com/', 'https://www.apple.com/'),
    ]
    async with _new_root_url_dialog_open() as (nud, project):
        last_focused = None  # type: Optional[wx.Window]
        for (url_input, without_www_url, with_www_url) in CASES:
            with subtests.test(url_input=url_input):
                with _urlopen_responding_with({
                        without_www_url: _UrlOpenHttpResponse(code=200, url=with_www_url)}):
                    last_focused = SetFocus(nud.url_field, last_focused)
                    nud.url_field.Value = url_input
                    
                    last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                    assertEqual(False, nud.url_field_focused)
                    assertEqual(True, nud.url_cleaner_spinner.IsShown())
                    await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
                    assertEqual(with_www_url, nud.url_field.Value)


@awith_subtests
async def test_given_url_input_is_nonempty_with_www_and_did_press_tab_and_spinner_is_visible_when_url_responds_with_http_3xx_to_url_without_www_and_url_without_www_responds_with_http_200_then_url_input_replaced_with_url_without_www_and_spinner_disappears(subtests: SubtestsContext) -> None:
    CASES = [
        ('www.xkcd.com/1/', 'https://www.xkcd.com/1/', 'https://xkcd.com/1/'),
        ('www.xkcd.com/', 'https://www.xkcd.com/', 'https://xkcd.com/'),
        ('www.xkcd.com', 'https://www.xkcd.com/', 'https://xkcd.com/'),
    ]
    async with _new_root_url_dialog_open() as (nud, project):
        last_focused = None  # type: Optional[wx.Window]
        for (url_input, with_www_url, without_www_url) in CASES:
            with subtests.test(url_input=url_input):
                with _urlopen_responding_with({
                        with_www_url: _UrlOpenHttpResponse(code=200, url=without_www_url)}):
                    last_focused = SetFocus(nud.url_field, last_focused)
                    nud.url_field.Value = url_input
                    
                    last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                    assertEqual(False, nud.url_field_focused)
                    assertEqual(True, nud.url_cleaner_spinner.IsShown())
                    await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
                    assertEqual(without_www_url, nud.url_field.Value)


@awith_subtests
async def test_given_url_input_is_nonempty_and_did_press_tab_and_spinner_is_visible_when_url_responds_with_http_3xx_to_unrelated_url_then_spinner_disappears(subtests: SubtestsContext) -> None:
    CASES = [
        ('contoso.com/', 'https://contoso.com/', 'https://www.microsoft.com/'),
    ]
    async with _new_root_url_dialog_open() as (nud, project):
        last_focused = None  # type: Optional[wx.Window]
        for (start_url, normalized_start_url, target_url) in CASES:
            with subtests.test(url_input=start_url):
                with _urlopen_responding_with({
                        start_url: _UrlOpenHttpResponse(code=200, url=target_url)}):
                    last_focused = SetFocus(nud.url_field, last_focused)
                    nud.url_field.Value = start_url
                    
                    last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                    assertEqual(False, nud.url_field_focused)
                    assertEqual(True, nud.url_cleaner_spinner.IsShown())
                    await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
                    assertEqual(normalized_start_url, nud.url_field.Value)


# === Test: Concurrent Actions While Resolving URL & Allow Create Root URL ===

async def test_given_url_input_is_unfocused_and_spinner_is_visible_when_focus_url_input_then_spinner_disappears() -> None:
    # TODO: Respond with "unreachable error" to be more realistic
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=500, url=ANY)):
        with _urlopen_paused():
            async with _new_root_url_dialog_open() as (nud, project):
                last_focused = None  # type: Optional[wx.Window]
                
                last_focused = SetFocus(nud.url_field, last_focused)
                nud.url_field.Value = '1.99.1.99'  # arbitrary IP
                
                last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                assertEqual(False, nud.url_field_focused)
                assertEqual(True, nud.url_cleaner_spinner.IsShown())
                
                last_focused = SetFocus(nud.url_field, last_focused)  # simulate press shift-tab
                assertEqual(True, nud.url_field_focused)
                await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
                
                nud.url_field.Value = '2.99.2.99'  # arbitrary different IP
                
                last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                assertEqual(False, nud.url_field_focused)
                assertEqual(True, nud.url_cleaner_spinner.IsShown())


async def test_given_url_input_is_nonempty_and_did_press_tab_and_spinner_is_visible_when_press_ok_then_disables_all_controls_except_cancel() -> None:
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=200, url=ANY)):
        async with _new_root_url_dialog_open() as (nud, project):
            last_focused = None  # type: Optional[wx.Window]
            
            with _urlopen_paused():
                last_focused = SetFocus(nud.url_field, last_focused)
                nud.url_field.Value = '1.99.1.99'  # arbitrary IP
                
                last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                assertEqual(False, nud.url_field_focused)
                assertEqual(True, nud.url_cleaner_spinner.IsShown())
                
                nud.name_field.Value = 'Home'
                
                click_button(nud.ok_button)
                assertEqual(False, nud.url_field.Enabled)
                assertEqual(False, nud.name_field.Enabled)
                assertEqual(False, nud.ok_button.Enabled)
                assertEqual(True, nud.cancel_button.Enabled)
            
            await wait_for(lambda: (False == nud.shown) or None)
            
            r = project.get_resource('https://1.99.1.99/')
            assert r is not None
            rr = project.get_root_resource(r)
            assert rr is not None
            assertEqual('Home', rr.name)


@skip('covered by: test_given_url_input_is_nonempty_and_did_press_tab_and_spinner_is_visible_when_press_ok_then_disables_all_controls_except_cancel')
async def test_given_url_input_is_nonempty_and_did_press_tab_and_spinner_is_visible_and_did_press_ok_when_spinner_disappears_then_dialog_disappears_and_root_url_is_created() -> None:
    pass


async def test_given_url_input_is_nonempty_and_did_press_tab_and_spinner_is_visible_and_did_press_ok_when_press_cancel_then_dialog_disappears() -> None:
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=200, url=ANY)):
        async with _new_root_url_dialog_open() as (nud, project):
            last_focused = None  # type: Optional[wx.Window]
            
            with _urlopen_paused():
                last_focused = SetFocus(nud.url_field, last_focused)
                nud.url_field.Value = '1.99.1.99'  # arbitrary IP
                
                last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                assertEqual(False, nud.url_field_focused)
                assertEqual(True, nud.url_cleaner_spinner.IsShown())
                
                click_button(nud.ok_button)
                assertEqual(False, nud.url_field.Enabled)
                assertEqual(False, nud.name_field.Enabled)
                assertEqual(False, nud.ok_button.Enabled)
                assertEqual(True, nud.cancel_button.Enabled)
                
                click_button(nud.cancel_button)
            
            await wait_for(lambda: (False == nud.shown) or None)
            
            r = project.get_resource('https://1.99.1.99/')
            assert r is None


async def test_given_url_input_is_unfocused_and_spinner_is_not_visible_when_press_ok_then_dialog_disappears_and_root_url_is_created() -> None:
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=200, url=ANY)):
        async with _new_root_url_dialog_open() as (nud, project):
            last_focused = None  # type: Optional[wx.Window]
            
            last_focused = SetFocus(nud.url_field, last_focused)
            nud.url_field.Value = 'xkcd.com'
            
            last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
            assertEqual(False, nud.url_field_focused)
            assertEqual(True, nud.url_cleaner_spinner.IsShown())
            
            nud.name_field.Value = 'Home'
            
            await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
            
            click_button(nud.ok_button)
            assertEqual(False, nud.shown)
            
            r = project.get_resource('https://xkcd.com/')
            assert r is not None
            rr = project.get_root_resource(r)
            assert rr is not None
            assertEqual('Home', rr.name)


async def test_given_url_input_is_focused_and_spinner_is_not_visible_when_press_ok_then_dialog_disappears_and_root_url_is_created() -> None:
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=200, url=ANY)):
        async with _new_root_url_dialog_open() as (nud, project):
            last_focused = None  # type: Optional[wx.Window]
            
            last_focused = SetFocus(nud.url_field, last_focused)
            nud.url_field.Value = 'xkcd.com'
            
            click_button(nud.ok_button)
            await wait_for(lambda: (False == nud.shown) or None)
            
            r = project.get_resource('https://xkcd.com/')
            assert r is not None
            rr = project.get_root_resource(r)
            assert rr is not None


async def test_given_url_input_is_focused_and_spinner_is_not_visible_when_press_cancel_then_dialog_disappears() -> None:
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=200, url=ANY)):
        async with _new_root_url_dialog_open() as (nud, project):
            last_focused = None  # type: Optional[wx.Window]
            
            last_focused = SetFocus(nud.url_field, last_focused)
            nud.url_field.Value = 'xkcd.com'
            
            click_button(nud.cancel_button)
            await wait_for(lambda: (False == nud.shown) or None)
            
            r = project.get_resource('https://xkcd.com/')
            assert r is None


async def test_given_url_input_is_unfocused_when_is_focused_and_is_unfocused_then_spinner_does_not_appear() -> None:
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=200, url=ANY)):
        async with _new_root_url_dialog_open() as (nud, project):
            last_focused = None  # type: Optional[wx.Window]
            
            last_focused = SetFocus(nud.url_field, last_focused)
            nud.url_field.Value = 'xkcd.com'
            
            last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
            assertEqual(False, nud.url_field_focused)
            assertEqual(True, nud.url_cleaner_spinner.IsShown())
            
            nud.name_field.Value = 'Home'
            
            await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
            assertEqual('https://xkcd.com/', nud.url_field.Value)
            
            last_focused = SetFocus(nud.url_field, last_focused)  # simulate press shift-tab
            
            last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
            assertEqual(False, nud.url_field_focused)
            assertEqual(False, nud.url_cleaner_spinner.IsShown())


# === Test: Disallow Create Empty Root URL ===

async def test_given_url_input_is_empty_then_ok_button_is_disabled() -> None:
    async with _new_root_url_dialog_open() as (nud, project):
        assertEqual('', nud.url_field.Value)
        assertEqual(False, nud.ok_button.Enabled)
        
        last_focused = None  # type: Optional[wx.Window]
        
        last_focused = SetFocus(nud.url_field, last_focused)
        nud.url_field.Value = 'xkcd.com'
        assertEqual(True, nud.ok_button.Enabled)
        
        last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
        assertEqual(True, nud.ok_button.Enabled)
        
        last_focused = SetFocus(nud.url_field, last_focused)  # simulate press shift-tab
        assertEqual(True, nud.ok_button.Enabled)
        
        nud.url_field.Value = ''
        assertEqual(False, nud.ok_button.Enabled)


@skip('covered by: test_given_url_input_is_empty_then_ok_button_is_disabled')
async def test_given_url_input_is_empty_when_url_input_becomes_nonempty_then_ok_button_is_enabled() -> None:
    pass


@skip('covered by: test_given_url_input_is_empty_then_ok_button_is_disabled')
async def test_given_url_input_is_nonempty_when_url_input_becomes_empty_then_ok_button_is_disabled() -> None:
    pass


# === Test: Disallow Create Duplicate Root URL ===

@awith_subtests
async def test_given_url_input_matches_existing_root_url_when_press_ok_then_displays_error_dialog_and_enables_all_controls(subtests: SubtestsContext) -> None:
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=200, url=ANY)):
        with subtests.test(case='given url input is focused'):
            async with _new_root_url_dialog_open() as (nud, project):
                r = Resource(project, 'https://xkcd.com/')
                RootResource(project, '', r)
                
                last_focused = None  # type: Optional[wx.Window]
                
                last_focused = SetFocus(nud.url_field, last_focused)
                nud.url_field.Value = 'xkcd.com/'
                
                with patch(
                        'crystal.browser.new_root_url.ShowModal',
                        mocked_show_modal('cr-root-url-exists', wx.ID_OK)
                        ) as show_modal_method:
                    click_button(nud.ok_button)
                    await wait_for(lambda: (1 == show_modal_method.call_count) or None)
                    await wait_for(lambda: (True == nud.url_field.Enabled) or None)
                assertEqual(True, nud.url_field.Enabled)
                assertEqual(True, nud.name_field.Enabled)
                assertEqual(True, nud.ok_button.Enabled)
                assertEqual(True, nud.cancel_button.Enabled)
        
        with subtests.test(case='given url input is unfocused and spinner is visible'):
            async with _new_root_url_dialog_open() as (nud, project):
                r = Resource(project, 'https://xkcd.com/')
                RootResource(project, '', r)
                
                last_focused = None
                
                last_focused = SetFocus(nud.url_field, last_focused)
                nud.url_field.Value = 'xkcd.com/'
                
                last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                assertEqual(False, nud.url_field_focused)
                assertEqual(True, nud.url_cleaner_spinner.IsShown())
                
                with patch(
                        'crystal.browser.new_root_url.ShowModal',
                        mocked_show_modal('cr-root-url-exists', wx.ID_OK)
                        ) as show_modal_method:
                    click_button(nud.ok_button)
                    await wait_for(lambda: (1 == show_modal_method.call_count) or None)
                    await wait_for(lambda: (True == nud.url_field.Enabled) or None)
                assertEqual(True, nud.url_field.Enabled)
                assertEqual(True, nud.name_field.Enabled)
                assertEqual(True, nud.ok_button.Enabled)
                assertEqual(True, nud.cancel_button.Enabled)
        
        with subtests.test(case='given url input is unfocused and spinner is not visible'):
            async with _new_root_url_dialog_open() as (nud, project):
                r = Resource(project, 'https://xkcd.com/')
                RootResource(project, '', r)
                
                last_focused = None
                
                last_focused = SetFocus(nud.url_field, last_focused)
                nud.url_field.Value = 'xkcd.com/'
                
                last_focused = SetFocus(nud.name_field, last_focused)  # simulate press tab
                assertEqual(False, nud.url_field_focused)
                assertEqual(True, nud.url_cleaner_spinner.IsShown())
                
                await wait_for(lambda: (False == nud.url_cleaner_spinner.IsShown()) or None)
                
                with patch(
                        'crystal.browser.new_root_url.ShowModal',
                        mocked_show_modal('cr-root-url-exists', wx.ID_OK)
                        ) as show_modal_method:
                    click_button(nud.ok_button)
                    await wait_for(lambda: (1 == show_modal_method.call_count) or None)
                    await wait_for(lambda: (True == nud.url_field.Enabled) or None)
                assertEqual(True, nud.url_field.Enabled)
                assertEqual(True, nud.name_field.Enabled)
                assertEqual(True, nud.ok_button.Enabled)
                assertEqual(True, nud.cancel_button.Enabled)


# === Utility ===

@contextmanager
def _EXPAND_enabled() -> Iterator[None]:
    old_value = os.environ.get('CRYSTAL_URLOPEN_MOCKED', 'False')
    os.environ['CRYSTAL_URLOPEN_MOCKED'] = 'True'
    try:
        yield
    finally:
        os.environ['CRYSTAL_URLOPEN_MOCKED'] = old_value


@asynccontextmanager
async def _new_root_url_dialog_open(*,
        autoclose: bool=True
        ) -> AsyncIterator[tuple[NewRootUrlDialog, Project]]:
    # Never allow automated tests to make real internet requests
    with _urlopen_responding_with(_UrlOpenHttpResponse(code=590, url=ANY)):
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            click_button(mw.new_root_url_button)
            nud = await NewRootUrlDialog.wait_for()
            nud.do_not_download_immediately()
            
            try:
                yield (nud, project)
            finally:
                if autoclose and nud.shown:
                    await nud.cancel()


@contextmanager
def _urlopen_responding_with(responses: 'Union[_UrlOpenHttpResponse, Dict[str, _UrlOpenHttpResponse]]') -> Iterator[None]:
    with _EXPAND_enabled():
        def mock_urlopen(url: str, timeout: float=0):
            if isinstance(responses, _UrlOpenHttpResponse):
                return responses
            else:
                try:
                    return responses[url]
                except IndexError:
                    return _UrlOpenHttpResponse(code=591, url=url)
        
        with patch('crystal.url_input.urlopen', mock_urlopen):
            yield


class _UrlOpenHttpResponse:
    def __init__(self, code: int, url: str) -> None:
        self._code = code
        self._url = url
    
    def __enter__(self) -> Self:
        return self
    
    def __exit__(self, *args) -> None:
        pass
    
    def getcode(self) -> int:
        return self._code
    
    def geturl(self) -> str:
        return self._url


@contextmanager
def _urlopen_paused() -> Iterator[None]:
    paused = True
    waiting = False
    done = False
    
    original_func = crystal.url_input.urlopen  # capture
    
    def pausable_urlopen(url: str, timeout: float=0):
        nonlocal done
        
        waiting = True
        while paused:
            time.sleep(DEFAULT_WAIT_PERIOD)
        
        result = original_func(url, timeout=timeout)
        
        waiting = False
        done = True
        return result
    
    with patch('crystal.url_input.urlopen', pausable_urlopen):
        yield
        paused = False
        if waiting:
            while not done:
                time.sleep(DEFAULT_WAIT_PERIOD)


def _assert_contains_sublist(xs: list[str], ys: list[str]) -> None:
    last_index = -1
    for y in ys:
        try:
            last_index = xs.index(y, last_index + 1)
        except IndexError:
            raise AssertionError(f'Expected list {xs} to contain sublist {ys} but it does not')


# NOTE: Only for use with tree items in EntityTree
_assert_tree_item_icon_tooltip_contains = EntityTree.assert_tree_item_icon_tooltip_contains
